"""Tests for complexity annotations"""

from typing import Type
import numpy as np
from numpy.lib.arraysetops import isin
import pytest

import pybryt
import pybryt.complexities as cplx

from pybryt.execution import MemoryFootprintValue

from .utils import assert_object_attrs


def generate_complexity_footprint(name, t_transform, max_exp=8):
    values = []
    for i, e in enumerate(range(1, max_exp + 1)):
        n = 10 ** e
        t = t_transform(n)
        values.append(MemoryFootprintValue(pybryt.TimeComplexityResult(name, n, 0, t), i, None))
    return pybryt.MemoryFootprint.from_values(*values)


def test_complexity_abc():
    """
    """
    pybryt.Annotation.reset_tracked_annotations()

    a =  pybryt.TimeComplexity(cplx.constant, name="foo")
    assert_object_attrs(a, {
        "name": "foo",
        "complexity": cplx.constant,
    })

    b =  pybryt.TimeComplexity(cplx.constant, name="foo")
    assert a == b

    b.complexity = cplx.linear
    assert a != b

    b.name = "bar"
    assert a != b

    b.complexity = cplx.constant
    assert a != b

    # test constructor errors
    with pytest.raises(ValueError, match="Complexity annotations require a 'name' kwarg"):
         pybryt.TimeComplexity(cplx.constant)

    with pytest.raises(ValueError, match="Invalid valid for argument 'complexity': 1"):
         pybryt.TimeComplexity(1, name="foo")


def test_time_complexity():
    pybryt.Annotation.reset_tracked_annotations()

    a =  pybryt.TimeComplexity(cplx.constant, name="foo")

    footprint = generate_complexity_footprint("foo", lambda v: 1012)
    res = a.check(footprint)
    assert res.satisfied
    assert res.value == cplx.constant

    footprint.add_value(np.random.uniform(size=100), 9)
    footprint.add_value( pybryt.TimeComplexityResult("bar", 10, 0, 10 ** 3), 10)
    res = a.check(footprint)
    assert res.satisfied
    assert res.value == cplx.constant

    footprint = generate_complexity_footprint("foo", np.log2)
    res = a.check(footprint)
    assert not res.satisfied
    assert res.value == cplx.logarithmic

    footprint = generate_complexity_footprint("foo", lambda v: v * np.log2(v))
    res = a.check(footprint)
    assert not res.satisfied
    assert res.value == cplx.linearithmic

    a.complexity = cplx.exponential
    res = a.check(footprint)
    assert not res.satisfied
    assert res.value == cplx.linearithmic


def test_alias():
    from pybryt.annotations.complexity import complexities as cplx2
    assert cplx.complexity_classes is cplx2.complexity_classes


def test_complexity_union():
    """
    Tests for complexity unions.
    """
    union = cplx.logarithmic | cplx.linear
    assert isinstance(union, cplx.ComplexityUnion)
    assert set(union.get_complexities()) == {cplx.logarithmic, cplx.linear}

    union2 = union | cplx.constant
    assert isinstance(union2, cplx.ComplexityUnion)
    assert set(union2.get_complexities()) == {cplx.logarithmic, cplx.linear, cplx.constant}

    union3 = cplx.constant | union
    assert isinstance(union3, cplx.ComplexityUnion)
    assert set(union3.get_complexities()) == {cplx.logarithmic, cplx.linear, cplx.constant}

    union4 = union2 | union
    assert isinstance(union4, cplx.ComplexityUnion)
    assert set(union4.get_complexities()) == {cplx.logarithmic, cplx.linear, cplx.constant}

    union.add_complexity(cplx.linearithmic)
    assert set(union.get_complexities()) == {cplx.logarithmic, cplx.linear, cplx.linearithmic}

    assert union == cplx.logarithmic | cplx.linear | cplx.linearithmic
    assert union != cplx.logarithmic | cplx.linear

    # test whether the union works with annotations
    a = pybryt.TimeComplexity(cplx.constant | cplx.logarithmic, name="foo")
    footprint = generate_complexity_footprint("foo", lambda v: 1012)
    res = a.check(footprint)
    assert res.satisfied
    assert res.value == cplx.constant

    footprint = generate_complexity_footprint("foo", lambda v: v)
    res = a.check(footprint)
    assert not res.satisfied
    assert res.value == cplx.linear

    # test errors
    with pytest.raises(TypeError):
        cplx.constant | 1

    with pytest.raises(TypeError):
        cplx.ComplexityUnion.from_or(1, cplx.constant)



"""Submission preprocessors for PyBryt"""

import nbformat

from typing import List, Set

from .abstract_preprocessor import AbstractPreprocessor
from .imports import ImportFindingPreprocessor
from .intermediate_variables import IntermediateVariablePreprocessor


PREPROCESSORS = [
    ImportFindingPreprocessor,
    IntermediateVariablePreprocessor
]


class NotebookPreprocessor(AbstractPreprocessor):
    """
    A class for applying a series of preprocessors to a notebook.
    """

    preprocessors: List[AbstractPreprocessor]

    def __init__(self) -> None:
        self.preprocessors = []
        for preprocessor_class in PREPROCESSORS:
            self.preprocessors.append(preprocessor_class())

    def preprocess(self, nb: nbformat.NotebookNode) -> nbformat.NotebookNode:
        for preprocessor in self.preprocessors:
            nb = preprocessor.preprocess(nb)

        return nb

    def get_imports(self) -> Set[str]:
        """
        Get the set of imports from the :py:class:`ImportFinder<pybryt.preprocessors.imports.ImportFindingPreprocessor`.

        Returns:
            ``set[str]``: the set of modules imported
        """
        import_finder_idx = [isinstance(p, ImportFindingPreprocessor) for p in self.preprocessors].index(True)
        return self.preprocessors[import_finder_idx].imports



"""Annotations for asserting the presence of a value"""

__all__ = ["Value", "Attribute", "ReturnValue"]

import dill
import numbers
import numpy as np
import pandas as pd

from collections.abc import Iterable, Sized
from copy import copy
from typing import Any, Callable, Dict, List, Optional, Union

from .annotation import Annotation, AnnotationResult
from .initial_condition import InitialCondition
from .invariants import invariant

from ..debug import _debug_mode_enabled
from ..execution import Event, MemoryFootprint, MemoryFootprintValue


class Value(Annotation):
    """
    Annotation class for asserting that a value should be observed.

    Indicates that a value passed to the constructor should be observed while tracing through the
    students' code. Values can be of any type that is pickleable by ``dill``. Values can specify a
    list of :ref:`invariants<invariants>` that will allow objects to be considered "equal." For 
    values that support arithmetic operators, absolute tolerances can be specified as well.

    Numeric tolerances are computed as with ``numpy.allcose``, where the value is considered "equal 
    enough" if it is within :math:`v \\pm (\\texttt{atol} + \\texttt{rtol} \\cdot |v|)`, where 
    :math:`v` is the value of the annotation.

    Args:
        value (``object``): the value that should be observed
        atol (``float`` or ``int``, optional): absolute tolerance for numeric values
        rtol (``float`` or ``int``, optional): relative tolerance for numeric values
        invariants (``list[invariant]``): invariants for 
            this value
        equivalence_fn (``callable[[object, object], bool]``): an optional function to check for
            equivalence between two values, overriding the default provided by ``Value``
        **kwargs: additional keyword arguments passed to the 
            :py:class:`Annotation<pybryt.annotations.annotation.Annotation>` constructor
    """

    value: Any
    """a copy of the value passed to the constructor"""

    atol: Optional[Union[float, int]]
    """absolute tolerance for numeric values"""

    rtol: Optional[Union[float, int]]
    """relative tolerance for numeric values"""

    invariants: List[invariant]
    """the invariants for this value"""

    equivalence_fn: Optional[Callable[[Any, Any], bool]]
    """
    a function that compares two values and returns True if they're "equal enough" and False
    otherwise
    """

    def __init__(
        self, 
        value: Any, 
        atol: Optional[Union[float, int]] = None, 
        rtol: Optional[Union[float, int]] = None, 
        invariants: List[invariant] = [], 
        equivalence_fn: Optional[Callable[[Any, Any], bool]] = None,
        **kwargs,
    ):
        try:
            dill.dumps(value)
        except Exception as e:
            raise ValueError(f"Values must be serializable but the following error was thrown during serialization:\n{e}")

        if _debug_mode_enabled() and equivalence_fn is not None and (atol is not None or rtol is not None):
            raise ValueError("Absolute or relative tolerance specified with an equivalence function")

        self.value = copy(value)
        self.atol = atol
        self.rtol = rtol
        self.invariants = invariants
        self.equivalence_fn = equivalence_fn

        # check that there aren't any issues applying the invariants
        if not self._tracking_initial_condition:
            self._apply_invariants(self.value)

        super().__init__(**kwargs)

    @property
    def children(self) -> List[Annotation]:
        return []

    @property
    def _tracking_initial_condition(self) -> bool:
        """
        Whether this value is tracking an initial condition.
        """
        return isinstance(self.value, InitialCondition)

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts this annotation's details to a JSON-friendly dictionary format.

        Output dictionary contains the annotation's name, group, limit number, success message, and
        failure message, as well as an ``invariants`` key with a list of the names of all invariants
        used in this value annotation and a ``atol`` key with this annotation's tolerance. This 
        dictionary does *not* contain the value being tracked.

        Returns:
            ``dict[str, object]``: the dictionary representation of this annotation
        """
        d = super().to_dict()
        d.update({
            "invariants": [inv.__name__ for inv in self.invariants],
            "atol": self.atol,
            "rtol": self.rtol,
        })
        return d

    def _apply_invariants(self, value: Any) -> List[Any]:
        """
        Apply the invariants on this annotation to the provided value.

        Args:
            value (``object``): the initial value

        Returns:
            ``list[object]``: the values considered equal to the initial value under the invariants
        """
        values = [value]
        for inv in self.invariants:
            values = inv(values)
        return values

    def check(self, footprint: MemoryFootprint) -> AnnotationResult:
        """
        Checks that the value tracked by this annotation occurs in the memory footprint.

        Args:
            footprint (:py:class:`pybryt.execution.memory_footprint.MemoryFootprint`): the
                memory footprint to check against
        
        Returns:
            :py:class:`AnnotationResult`: the results of this annotation against ``footprint``
        """
        satisfier = self._get_satisfying_index(footprint)
        return self._generate_annotation_result(footprint, satisfier)

    def _get_satisfying_index(self, footprint: MemoryFootprint) -> Optional[int]:
        """
        Return the index of the first value in the memory footprint that satisfies this annotation.

        Args:
            footprint (:py:class:`pybryt.execution.memory_footprint.MemoryFootprint`): the
                memory footprint to check against

        Returns:
            ``int | None``: the index, or ``None`` if no value satisfies this annotation.
        """
        expected_value = self.value
        if isinstance(expected_value, InitialCondition):
            expected_value = expected_value.supply_footprint(footprint)
        satisfied = [self._check_observed_value(expected_value, mfp_val.value) for mfp_val in footprint]
        return satisfied.index(True) if any(satisfied) else None

    def _generate_annotation_result(
        self,
        footprint: MemoryFootprint,
        satisfying_index: Optional[int],
    ) -> AnnotationResult:
        """
        Return an ``AnnotationResult`` based on the provided satisfying index.

        Args:
            footprint (``MemoryFootprint``): the footprint being checked
            satisfying_index (``int | None``): the index of the value in the memory footprint that
                satisfies the annotation, otherwise ``None``

        Returns:
            ``AnnotationResult``: the result
        """
        if satisfying_index is None:
            return AnnotationResult(False, self)

        return AnnotationResult(
            True, 
            self, 
            footprint.get_value(satisfying_index).value,
            footprint.get_value(satisfying_index).timestamp,
        )

    def __eq__(self, other: Any) -> bool:
        """
        Checks whether this annotation is equal to another object.

        To be equal to a ``Value``, the other object must also be a ``Value`` object, have the same
        set of invariants, the same tolerance, and the same initial value.

        Args:
            other (``object``): the object to compare to

        Returns:
            ``bool``: whether the objects are equal
        """
        if not super().__eq__(other):
            return False

        if self._tracking_initial_condition:
            vals_equal = self.value == other.value
        else:
            vals_equal = self.check_values_equal(self.value, other.value)

        return self.invariants == other.invariants and vals_equal and \
            self.atol == other.atol and self.rtol == other.rtol and \
            self.equivalence_fn == self.equivalence_fn

    def check_against(self, other_value: Any) -> bool:
        """
        Check whether an object satisfies this annotation.

        Args:
            other_value (``object``): the value to check against

        Returns:
            ``bool``: whether this annotation is satisfied by the provided value
        """
        if self._tracking_initial_condition:
            raise ValueError("check_against cannot be used with initial conditions")

        return self.check(MemoryFootprint.from_values(MemoryFootprintValue(other_value, 0, None))).satisfied

    def _check_observed_value(self, expected_value: Any, observed_value: Any) -> bool:
        """
        Checks whether a single observed value tuple satisfies this value.

        Applies all invariants to the first element of the tuple and then checks whether any of the
        resulting values match and of the values in ``self._values``.

        Args:
            observed_value (``object``): the observed value

        Returns:
            ``bool``: whether the value matched
        """
        other_values = self._apply_invariants(observed_value)

        for value in self._apply_invariants(expected_value):
            for other_value in other_values:
                if self.check_values_equal(value, other_value, self.atol, self.rtol, self.equivalence_fn):
                    return True

        return False

    @staticmethod
    def check_values_equal(value, other_value, atol = None, rtol = None, equivalence_fn = None) -> bool:
        """
        Checks whether two objects are equal. 
        
        If the values are both numeric (numerics, arrays, etc.) and ``atol`` and/or ``rtol`` are 
        specified, the values are considered equal if ``other_value`` is within the tolerance
        bounds of ``value``.

        The equivalence check provided by this function can be overridden by providing a custom
        function to check equivalence. If provided, the value returned by this function is returned,
        unless an error is thrown, in which case ``False`` is returned.

        Args:
            value (``object``): the first object to compare
            other_value (``object``): the second object to compare
            atol (``float``, optional): the absolute tolerance for numeric values
            rtol (``float``, optional): the relative tolerance for numeric values
            equivalence_fn (``callable[[object, object], bool]``): an optional function to check 
                for equivalence between two values, overriding the default provided by ``Value`` 
        """
        if equivalence_fn is not None:
            try:
                ret = equivalence_fn(value, other_value)

            except:
                if _debug_mode_enabled():
                    raise

                return False

            if not isinstance(ret, bool):
                raise TypeError(f"Custom equivalence function returned value of invalid type: {type(ret)}")

            return ret

        if isinstance(value, Iterable) ^ isinstance(other_value, Iterable):
            return False

        if value is None:
            return other_value is None

        if atol is None:
            atol = 0

        if rtol is None:
            rtol = 0

        if hasattr(value, 'shape'):
            try:
                if value.shape != other_value.shape:
                    return False
            except AttributeError:
                return False

        try:
            ub, lb = value + atol, value - atol
            ub += rtol * np.abs(value)
            lb -= rtol * np.abs(value)
            numeric = True

        except:
            numeric = False

        if numeric:
            try:
                res = np.logical_and(ub >= other_value, other_value >= lb)
            except (ValueError, TypeError) as e:
                return False
        else:
            try:
                if (hasattr(value, "shape") and hasattr(other_value, "shape") \
                        and value.shape != other_value.shape) \
                        or (hasattr(value, "shape") ^ hasattr(other_value, "shape")):
                    return False

                # tolerances make sense only for iterables with numerical data
                if isinstance(value, Sized) and len(value) > 0 and isinstance(value, Iterable) \
                        and all(isinstance(i, numbers.Real) for i in value):
                    # np.allclose doesn't work with sets
                    if isinstance(value, set):
                        if len(value) != len(other_value):
                            return False
                        res = np.array(np.isclose(v, o, atol=atol, rtol=rtol) for v, o in zip(sorted(value), sorted(other_value))).all()
                    elif isinstance(value, dict):
                        if not isinstance(other_value, dict):
                            return False
                        if all(isinstance(i, numbers.Real) for i in value.values()):
                            res = np.array(np.isclose(k1, k2) and np.isclose(v1, v2) \
                                for (k1, v1), (k2, v2) in zip(sorted(value.items(), key=lambda t: t[0]), \
                                sorted(other_value.items(), key=lambda t: t[0]))).all()
                        else:
                            res = value == other_value
                    else:
                        res = np.allclose(value, other_value, atol=atol, rtol=rtol)

                else:
                    res = value == other_value

            except (ValueError, TypeError) as e:
                try:
                    res = value == other_value
                except:
                    if _debug_mode_enabled():
                        raise
                    return False

        if isinstance(res, Iterable):
            # handle NaN values in pandas objects
            if isinstance(res, (pd.DataFrame, pd.Series)) and hasattr(value, "isna") and \
                    hasattr(other_value, "isna"):
                res = res | (value.isna() & other_value.isna())
            if isinstance(res, (np.ndarray, pd.DataFrame, pd.Series)):
                res = res.all(axis=None)
            else:
                res = all(res)

        if res:
            return True

        return False


class _AttrValue(Value):
    """
    Wrapper around ``Value`` for checking whether an object has an *attribute* matching a specific
    value.

    Args:
        obj (``object``): the object being checked
        attr (``str``): the attribute of the object being checked
        enforce_type (``bool``, optional): whether to ensure that the satisfying value has the same 
            type as the initial value
        **kwargs: additional keyword arguments passed to the :py:class:`Value<pybryt.Value>` 
            constructor
    """

    _object: Any
    _attr: str

    enforce_type: bool
    """whether to ensure that the satisfying value has the same type as the initial value"""

    def __init__(self, obj: Any, attr: str, enforce_type: bool = False, **kwargs):
        self._object = obj
        self._attr = attr
        self.enforce_type = enforce_type
        val = getattr(obj, attr)
        super().__init__(val, **kwargs)
    
    def __eq__(self, other: Any) -> bool:
        """
        Checks whether this annotation is equal to another object.

        To be equal to a ``_AttrValue``, the other object must also be a ``_AttrValue`` object, have
        equal underlying ``Value`` annotations, have equal source objects, and have the same 
        ``_attr``.

        Args:
            other (``object``): the object to compare to

        Returns:
            ``bool``: whether the objects are equal
        """
        return super().__eq__(other) and self.check_values_equal(self._object, other._object) and \
            self._attr == other._attr and self.enforce_type == other.enforce_type
    
    def check(self, footprint: MemoryFootprint) -> AnnotationResult:
        """
        Checks whether any of the values in the memory footprint has an attribute matching the value
        in this annotation.

        Args:
            footprint (:py:class:`pybryt.execution.memory_footprint.MemoryFootprint`): the
                memory footprint to check against
        
        Returns:
            :py:class:`AnnotationResult`: the results of this annotation against ``footprint``
        """
        orig_mfp_vals, attr_mfp_vals = [], []
        for mfp_val in footprint:
            if not self.enforce_type or isinstance(mfp_val.value, type(self._object)):
                mfp_lst = mfp_val.to_list()
                if hasattr(mfp_val.value, self._attr):
                    mfp_lst[0] = getattr(mfp_val.value, self._attr)
                    attr_mfp_vals.append(MemoryFootprintValue(*mfp_lst))
                    orig_mfp_vals.append(mfp_val)

        attrs_fp = MemoryFootprint.from_values(*attr_mfp_vals)
        satisfying_index = self._get_satisfying_index(attrs_fp)
        child_res = self._generate_annotation_result(attrs_fp, satisfying_index)
        satisfier = None if satisfying_index is None else orig_mfp_vals[satisfying_index].value
        return AnnotationResult(None, self, value=satisfier, children=[child_res])


class Attribute(Annotation):
    """
    Annotation for asserting that an object has an attribute value matching that of another object.

    Uses a :py:class:`Value<pybryt.Value>` annotation to check there exists an object in the 
    students' code which as a specific attribute, or a series of attributes, with values matching
    those in the object passed to the constructor. The constructor taks in the object in question
    and the name of the attribute(s) being studied.

    Args:
        obj (``object``): the object being checked
        attrs (``str`` or ``list[str]``): the attribute or attributes that should be checked
        enforce_type (``bool``, optional): whether to ensure that the satisfying value has the same 
            type as the initial value
        **kwargs: additional keyword arguments passed to the 
            :py:class:`Annotation<pybryt.Annotation>` constructor
    """

    _annotations: List[_AttrValue]
    _invariants: List[invariant]
    _atol: Optional[Union[float, int]]
    _rtol: Optional[Union[float, int]]

    enforce_type: bool
    """whether to ensure that the satisfying value has the same type as the initial value"""

    def __init__(self, obj: Any, attrs: Union[str, List[str]], enforce_type: bool = False, **kwargs):
        if isinstance(attrs, str):
            attrs = [attrs]
        if not isinstance(attrs, list) or not all(isinstance(a, str) for a in attrs):
            raise TypeError(f"Invalid type for argument 'attrs': {type(attrs)}")
        if isinstance(obj, InitialCondition):
            raise TypeError("Initial conditions are not compatible with attribute annotations")

        name = kwargs.pop("name", None)
        success_message = kwargs.pop("success_message", None)
        failure_message = kwargs.pop("failure_message", None)
        self._annotations = []
        for attr in attrs:
            if not hasattr(obj, attr):
                raise AttributeError(f"{obj} has not attribute '{attr}'")
            self._annotations.append(_AttrValue(obj, attr, enforce_type=enforce_type, **kwargs))
        
        self._invariants = kwargs.pop("invariants", [])
        self._atol = kwargs.pop("atol", None)
        self._rtol = kwargs.pop("rtol", None)

        self.enforce_type = enforce_type
    
        super().__init__(
            name=name,
            success_message=success_message,
            failure_message=failure_message,
            **kwargs,
        )

    @property
    def children(self) -> List[Annotation]:
        return self._annotations
    
    def __eq__(self, other: Any) -> bool:
        """
        Checks whether this annotation is equal to another object.

        To be equal to an ``Attribute``, the other object must also be an ``Attribute`` object and
        have equal child annotations.

        Args:
            other (``object``): the object to compare to

        Returns:
            ``bool``: whether the objects are equal
        """
        return super().__eq__(other) and self.children == other.children

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts this annotation's details to a JSON-friendly dictionary format.

        Output dictionary contains the annotation's name, group, limit number, success message, and
        failure message, as well as an ``invariants`` key with a list of the names of all invariants
        used in this value annotation, a ``atol`` key with this annotation's tolerance, and a key for 
        the attributes being checked. This dictionary does *not* contain the value being tracked.

        Returns:
            ``dict[str, object]``: the dictionary representation of this annotation
        """
        d = super().to_dict()
        d.update({
            "invariants": [inv.__name__ for inv in self._invariants],
            "atol": self._atol,
            "rtol": self._rtol,
            "attributes": [av._attr for av in self._annotations],
            "enforce_type": self.enforce_type,
        })
        return d

    def check(self, footprint: MemoryFootprint) -> AnnotationResult:
        """
        Checks whether any of the values in the memory footprint has all of the required attributes,
        each matching the values expected.

        Args:
            footprint (:py:class:`pybryt.execution.memory_footprint.MemoryFootprint`): the
                memory footprint to check against
        
        Returns:
            :py:class:`AnnotationResult`: the results of this annotation against ``footprint``
        """
        results = [v.check(footprint) for v in self._annotations]        
        return AnnotationResult(None, self, children=results)

    def check_against(self, other_value: Any) -> bool:
        """
        Check whether an object satisfies this annotation.

        Args:
            other_value (``object``): the value to check against

        Returns:
            ``bool``: whether this annotation is satisfied by the provided value
        """
        return self.check(MemoryFootprint.from_values(MemoryFootprintValue(other_value, 0, None))).satisfied


class ReturnValue(Value):
    """
    Annotation class for asserting that a value should be returned by a student-written function.

    Indicates that a value passed to the constructor should be returned by a call to a student's
    function. Values can be of any type that is pickleable by ``dill``. Values can specify a
    list of :ref:`invariants<invariants>` that will allow objects to be considered "equal." For 
    values that support arithmetic operators, absolute tolerances can be specified as well.

    Numeric tolerances are computed as with ``numpy.allcose``, where the value is considered "equal 
    enough" if it is within :math:`v \\pm (\\texttt{atol} + \\texttt{rtol} \\cdot |v|)`, where 
    :math:`v` is the value of the annotation.

    Args:
        value (``object``): the value that should be observed
        atol (``float`` or ``int``, optional): absolute tolerance for numeric values
        rtol (``float`` or ``int``, optional): relative tolerance for numeric values
        invariants (``list[invariant]``): invariants for 
            this value
        equivalence_fn (``callable[[object, object], bool]``): an optional function to check for
            equivalence between two values, overriding the default provided by ``Value``
        **kwargs: additional keyword arguments passed to the 
            :py:class:`Annotation<pybryt.annotations.annotation.Annotation>` constructor
    """

    _VALID_EVENTS = {Event.RETURN, Event.LINE_AND_RETURN}

    def check(self, footprint: MemoryFootprint) -> AnnotationResult:
        satisfier = self._get_satisfying_index(footprint)
        if satisfier is not None and footprint.get_value(satisfier).event not in type(self)._VALID_EVENTS:
            satisfier = None
        return self._generate_annotation_result(footprint, satisfier)



"""Tests for abstract annotations and annotation results"""

import pybryt
from pybryt.execution.memory_footprint import MemoryFootprint

from .utils import assert_object_attrs, generate_memory_footprint


def test_name_group_limit():
    """
    """
    footprint = generate_memory_footprint()
    val = footprint.get_value(2)

    pybryt.Annotation.reset_tracked_annotations()
    vs = []
    for _ in range(100):
        vs.append(pybryt.Value(val, name="foo", limit=11))
    
    tracked = pybryt.Annotation.get_tracked_annotations()
    assert len(tracked) == 100
    assert tracked == vs, "Wrong tracked annotations"

    res = vs[-1].check(footprint)
    assert_object_attrs(res, {
        "name": "foo",
        "group": None,
    })

    v1 = pybryt.Value(footprint.get_value(0).value, group="bar")
    v2 = pybryt.Value(footprint.get_value(1).value, group="bar")
    assert_object_attrs(v1, {"group": "bar"})
    assert_object_attrs(v2, {"group": "bar"})

    res = v1.check(footprint)
    print(res.name)
    assert_object_attrs(res, {
        "name": "Annotation 101",
        "group": "bar",
    })


def test_get_reset_tracked_annotations():
    """
    """
    tracked = pybryt.Annotation.get_tracked_annotations()
    pybryt.Annotation.reset_tracked_annotations()
    assert len(tracked) == 0

    v1 = pybryt.Value(1)
    v2 = pybryt.Value(2)
    assert len(tracked) == 2

    v3 = pybryt.Value(3)
    assert len(tracked) == 3

    v4 = v3.before(v2)
    assert len(tracked) == 2
    
    assert v1 in tracked
    assert v2 not in tracked
    assert v3 not in tracked
    assert v4 in tracked

    pybryt.Annotation.reset_tracked_annotations()
    assert len(tracked) == 0


def test_messages():
    """
    """
    footprint = generate_memory_footprint()
    pybryt.Annotation.reset_tracked_annotations()

    val1 = footprint.get_value(0).value
    val2 = footprint.get_value(1).value

    v1 = pybryt.Value(val1, success_message="m1", failure_message="m2")
    v2 = pybryt.Value(val2)
    
    v = v1.before(v2)
    res = v.check(footprint)

    assert len(res.messages) == 3
    assert_object_attrs(res.messages[0], {"message": "m1", "satisfied": True})

    v.failure_message = "m3"
    res = v.check(footprint)

    assert len(res.messages) == 3

    res = v.check(MemoryFootprint.from_values())

    assert len(res.messages) == 3
    assert_object_attrs(res.messages[0], {"message": "m2", "satisfied": False})
    assert_object_attrs(res.messages[1], {"message": None, "satisfied": False})
    assert_object_attrs(res.messages[2], {"message": "m3", "satisfied": False})

    v2.name = "v2"
    v2.failure_message = "m4"
    res = v.check(MemoryFootprint.from_values())

    assert len(res.messages) == 3
    assert_object_attrs(res.messages[1], {"message": "m4", "name": "v2", "satisfied": False})


def test_bitwise_ops():
    """
    """
    a1, a2 = pybryt.Value(1), pybryt.Value(2)

    assert isinstance(a1 & a2, pybryt.AndAnnotation), "& operator returns wrong type"
    assert isinstance(a1 | a2, pybryt.OrAnnotation), "| operator returns wrong type"
    assert isinstance(a1 ^ a2, pybryt.XorAnnotation), "^ operator returns wrong type"
    assert isinstance(~a1, pybryt.NotAnnotation), "~ operator returns wrong type"



"""Memory footprint container for PyBryt"""

import nbformat

from dataclasses import astuple, dataclass
from enum import Enum
from typing import Any, Dict, List, Optional, Set, Tuple, Union

from ..utils import filter_pickleable_list, pickle_and_hash


class Event(Enum):
    """
    An enum of event ``sys.settrace`` event types traced by PyBryt.
    """

    LINE = "line"
    """indicates that a value was seen in a "line" event"""

    RETURN = "return"
    """indicates that a value was seen in a "return" event"""

    LINE_AND_RETURN = "line_and_return"
    """indicates that a value was seen in both a "line" and "return" event"""

    @classmethod
    def from_event_name(cls, event_name: str) -> Optional["Event"]:
        """
        Return the enum value corresponding to the event name, if present.

        Args:
            event_name (``str``): the event name provided by ``sys.settrace``

        Returns:
            :py:class:`Event`: the enum value corresponding to the event name if present,
                otherwise ``None``
        """
        return cls(event_name) if event_name in {v.value for v in cls.__members__.values()} else None


class Counter:
    """
    A counter for tacking the step number in the tracing function.

    Args:
        start (``int``): the starting value of the counter
    """

    val: int
    """the current value of the counter"""

    def __init__(self, start: int = 0):
        self.val = start

    def get_value(self) -> int:
        """
        Get the current value of the counter.

        Returns:
            ``int``: the current value of the counter
        """
        return self.val

    def increment(self) -> None:
        """
        Increment the counter by one.
        """
        self.val += 1

    def offset(self, val: int) -> None:
        """
        Offset the counter by a specific amount.

        Args:
            val (``int``): the amount to offset by
        """
        self.val += val


@dataclass
class MemoryFootprintValue:
    """
    A data class for an observed value in the memory footprint.

    The fields and their order should correspond to the structure of the tuples in
    :py:attribute:`MemoryFootprint.values`.
    """

    value: Any
    """the value"""

    timestamp: int
    """the timestamp at which this value was observed"""

    event: Optional[Event]
    """the trace function event that created this value, if applicable"""

    def to_list(self):
        return list(astuple(self))


class MemoryFootprint:
    """
    A memory footprint for an executed notebook.

    Args:
        counter (:py:class:`pybryt.execution.memory_footprint.Counter`, optional): a counter to use 
            for this footprint; if unprovided, a new one is initialized
    """

    counter: Counter
    """the counter used to construct this footprint"""

    _value_indices_by_hash: Dict[str, int]
    """indices of values keyed on their hashes"""

    values: List[Tuple[Any, int, Optional[Event]]]
    """the values, timestamps, and event types in the footprint"""

    calls: List[Tuple[str, str]]
    """the list of function calls tracked by the trace function"""

    imports: Set[str]
    """the set of modules imported during execution"""

    executed_notebook: Optional[nbformat.NotebookNode]
    """the final (pre-processed) notebook that was executed, with outputs"""

    initial_conditions: Dict[str, Any]
    """initial conditions set during execution"""

    def __init__(self, counter: Optional[Counter] = None):
        self.counter = counter if counter is not None else Counter()
        self._value_indices_by_hash = {}
        self.values = []
        self.calls = []
        self.imports = set()
        self.executed_notebook = None
        self.initial_conditions = {}

    @classmethod
    def from_values(cls, *values: MemoryFootprintValue) -> 'MemoryFootprint':
        """
        Generate a memory footprint from values and their timestamps.
        Args:
            values_and_timestamps (:py:class:`MemoryFootprintValue`): the values

        Returns:
            :py:class:`pybryt.execution.memory_footprint.MemoryFootprint`: the memory footprint

        Raises:
            ``TypeError``: if any of the values is of the wrong type
        """
        if not all(isinstance(v, MemoryFootprintValue) for v in values):
            raise TypeError("Arguments to from_values must be of type MemoryFootprintValue")

        values = [astuple(v) for v in values]

        footprint = cls()
        footprint.values.extend(values)
        footprint.offset_counter(footprint.num_steps)
        return footprint

    @classmethod
    def combine(cls, *footprints: 'MemoryFootprint') -> 'MemoryFootprint':
        """
        Generate a memory footprint by combining memory footprints in-sequence.

        Duplicate values in different footprints are removed, keeping the first instance seen. The
        timestmaps of each footprint are offset by the number of steps in all preceding footprints.

        Args:
            *footprints (:py:class:`pybryt.execution.memory_footprint.MemoryFootprint`): the 
                footprints

        Returns:
            :py:class:`pybryt.execution.memory_footprint.MemoryFootprint`: the memory footprint
        """
        new_fp = cls()
        seen = set()          # set to track which values we've seen
        timestamp_offset = 0  # offset for timestamps in the new memory footprint
        for fp in footprints:
            map(lambda c: new_fp.add_call(*c), fp.calls)
            for fp_val in fp:
                h = pickle_and_hash(fp_val.value)
                if h not in seen:
                    fp_val.timestamp += timestamp_offset
                    new_fp.add_value(
                        fp_val.value,
                        fp_val.timestamp,
                        fp_val.event,
                        allow_duplicates=True,
                    )
                    seen.add(h)

            timestamp_offset += fp.num_steps

        new_fp.offset_counter(timestamp_offset)
        return new_fp

    def increment_counter(self) -> None:
        """
        Increment the step counter by one.
        """
        self.counter.increment()

    def offset_counter(self, val: int) -> None:
        """
        Offset the step counter by a specific amount.

        Args:
            val (``int``): the amount to offset by
        """
        self.counter.offset(val)

    def add_value(
        self,
        val: Any,
        timestamp: Optional[int] = None,
        event: Optional[Event] = None,
        allow_duplicates: bool = False,
    ) -> None:
        """
        Add a value to the memory footprint.

        If the timestamp is unspeficied, the step counter is polled for the current value. By default,
        this method does not allow duplicate values to be entered into the footprint; this can be
        disabled using ``allow_duplicates``.

        Args:
            value (``object``): the value to add
            timestamp (``int``, optional): the timestamp
            event (:py:class:`Event`, optional): the event that produced the value
            allow_duplicates(``bool``): whether duplicate values should be allowed in the footprint
        """
        if not allow_duplicates:
            h = pickle_and_hash(val)
            if h in self._value_indices_by_hash:
                tup = self.values[self._value_indices_by_hash[h]]
                tup_event = tup[2]

                # update the event type if necessary
                if tup_event is None and event is not None:
                    tup_event = event
                elif tup_event is not None and event is not None and tup_event != event:
                    tup_event = Event.LINE_AND_RETURN
                
                self.values[self._value_indices_by_hash[h]] = (tup[0], tup[1], tup_event)

                return

            self._value_indices_by_hash[h] = len(self.values)

        if timestamp is None:
            timestamp = self.counter.get_value()

        self.values.append((val, timestamp, event))

    def get_value(self, index: int) -> MemoryFootprintValue: # TODO: change name??
        """
        Get the value at the specified index.

        Args:
            index (``int``): the index

        Returns:
            :py:class:`MemoryFootprintValue`: the value
        """
        return MemoryFootprintValue(*self.values[index])

    def add_call(self, filename: str, fn_name: str) -> None:
        """
        Add a function call.

        Args:
            filename (``str``): the filename of the function
            fn_name (``str``): the name of the function
        """
        self.calls.append((filename, fn_name))

    def add_imports(self, *modules: str) -> None:
        """
        Add imports to the set of imports

        Args:
            *modules (``str``): the modules to add
        """
        self.imports.update(modules)

    def set_executed_notebook(self, nb: nbformat.NotebookNode) -> None:
        """
        Set the executed notebook of this memory footprint.

        Args:
            nb (``nbformat.NotebookNode``): the notebook
        """
        self.executed_notebook = nb

    def filter_out_unpickleable_values(self) -> None:
        """
        Filter any unpickleable objects out of the list of value-timestamp tuples in-place.
        """
        filter_pickleable_list(self.values)

    def clear(self) -> None:
        """
        """
        self.values.clear()

    def set_initial_conditions(self, initial_conditions: Dict[str, Any]) -> None:
        """
        Update the initial conditions with a dictionary.

        Args:
            initial_conditions (``dict[str, object]``): a dictionary of initial conditions
        """
        self.initial_conditions.update(initial_conditions)

    def get_initial_conditions(self) -> Dict[str, Any]:
        """
        Return the initial conditions dictionary.

        Returns:
            ``dict[str, object]``: the initial conditions dictionary
        """
        return self.initial_conditions

    @property
    def num_steps(self) -> int:
        """
        ``int``: the total number of steps this implementation took
        """
        timestamps = [t[1] for t in self.values]
        return max(timestamps) if len(timestamps) else -1

    def __len__(self):
        return len(self.values)

    def __iter__(self):
        return MemoryFootprintIterator(self)

    def __eq__(self, other: Any):
        """
        Return whether another object is equal to this one.

        An object is equal to a memory footprint if it is also a memory footprint and has the same
        values, calls, imports, and executed notebook.
        """
        return isinstance(other, type(self)) and self.calls == other.calls \
            and self.imports == other.imports \
            and self.executed_notebook == other.executed_notebook \
            and pickle_and_hash(self.values) == pickle_and_hash(other.values)


class MemoryFootprintIterator:
    """
    An iterator class for memory footprints.

    Args:
        footprint (:py:class:`MemoryFootprint`): the memory footprint to iterate over
        start (``int``, optional): the index to start at
    """

    footprint: MemoryFootprint
    """the memory footprint being iterated over"""

    _index: int
    """the next index in the memory footprint to yield"""

    def __init__(self, footprint: MemoryFootprint, start: int = 0):
        self.footprint = footprint
        self._index = start

    def __next__(self) -> MemoryFootprintValue:
        if self._index < len(self.footprint):
            value = self.footprint.get_value(self._index)
            self._index += 1
            return value
        raise StopIteration()



"""Auto-assessment library for Python"""

from . import utils
from .annotations import *
from .complexity import TimeComplexityChecker
from .debug import *
from .execution import *
from .reference import *
from .student import *
from .version import __version__



"""Tracing function and controls"""

import inspect
import linecache
import re

from copy import copy
from types import FrameType, FunctionType, ModuleType
from typing import Any, Dict, List, Optional, Tuple, Callable

from .complexity import is_complexity_tracing_enabled
from .memory_footprint import Event, MemoryFootprint
from .utils import is_ipython_frame

from ..utils import make_secret, pickle_and_hash, UnpickleableError


ACTIVE_FOOTPRINT = None
TRACING_FUNC = None
TRACING_VARNAME = "__PYBRYT_TRACING__"


def create_collector(
    skip_types: List[type] = [type, type(len), FunctionType],
    addl_filenames: List[str] = [],
) -> Tuple[MemoryFootprint, Callable[[FrameType, str, Any], Callable]]:
    """
    Creates a memory footprint to collect observed values and a trace function.

    Any types in ``skip_types`` won't be tracked by the trace function. The trace function by 
    default only traces inside IPython but can be set to trace inside specific files using the
    ``addl_filenames`` argument, which should be a list absolute paths to files that should also be
    traced inside of.

    Args:
        skip_types (``list[type]``, optional): object types not to track
        addl_filenames (``list[str]``, optional): filenames to trace inside of in addition to 
            IPython
        
    Returns:
        ``tuple[MemoryFootprint, callable[[frame, str, object], callable]]``: the memory footprint
        and the trace function
    """
    global ACTIVE_FOOTPRINT
    vars_not_found: Dict[str, List[Tuple[str, str, int]]] = {}
    footprint = MemoryFootprint()

    def track_value(val: Any, event_name: str, seen_at: Optional[int] = None):
        """
        Tracks a value in ``footprint``. Checks that the value has not already been tracked by 
        pickling it and hashing the pickled object and comparing it to ``hashes``. If pickling is
        unsuccessful, the value is not tracked.

        Args:
            val (``object``): the object to be tracked
            event_name (``str``): the event name provided by ``sys.settrace``
            seen_at (``int``, optional): an overriding step counter value
        """
        try:
            if hasattr(val, "__module__"):
                footprint.add_imports(val.__module__.split(".")[0])

            if type(val) in skip_types:
                return

            if isinstance(val, ModuleType):
                footprint.add_imports(val.__name__.split(".")[0])
                return

            event = Event.from_event_name(event_name)
            footprint.add_value(copy(val), seen_at, event)

        # if something fails, don't track
        except:
            return

    def track_call(frame):
        """
        Tracks a call in ``calls`` as a tuple of ``(filename, function name)``.

        Args:
            frame (``types.FrameType``): the frame of the call
        """
        footprint.add_call(frame.f_code.co_filename, frame.f_code.co_name)

    # TODO: a way to track the cell of execution
    def collect_intermidiate_results(frame: FrameType, event: str, arg: Any):
        """
        Trace function for PyBryt.
        """
        if is_ipython_frame(frame) or frame.f_code.co_filename in addl_filenames:
            footprint.increment_counter()  # increment student code step counter

        if event == "call":
            track_call(frame)
            return collect_intermidiate_results

        # return if tracking is disabled by a compelxity check
        if is_complexity_tracing_enabled():
            return collect_intermidiate_results

        name = frame.f_code.co_filename + frame.f_code.co_name

        if is_ipython_frame(frame) or frame.f_code.co_filename in addl_filenames:
            if event == "line" or event == "return":

                line = linecache.getline(frame.f_code.co_filename, frame.f_lineno)
                tokens = set("".join(char if char.isalnum() or char == '_' else "\n" for char in line).split("\n"))
                for t in "".join(char if char.isalnum() or char == '_' or char == '.' else "\n" for char in line).split("\n"):
                    tokens.add(t)
                tokens = sorted(tokens)  # sort for stable ordering
                
                for t in tokens:
                    if "." in t:
                        try:
                            float(t)  # prevent adding floats prematurely
                            continue
                        except ValueError:
                            pass

                        try:
                            val = eval(t, frame.f_globals, frame.f_locals)
                            track_value(val, event)
                        except:
                            pass

                    else:
                        if t in frame.f_locals:
                            val = frame.f_locals[t]
                            track_value(val, event)
                                
                        elif t in frame.f_globals:
                            val = frame.f_globals[t]
                            track_value(val, event)
                
                # for tracking the results of an assignment statement
                m = re.match(r"^\s*(\w+)(\[[^\]]\]|(\.\w+)+)*\s=.*", line)
                if m:
                    if name not in vars_not_found:
                        vars_not_found[name] = []
                    vars_not_found[name].append((m.group(1), event, footprint.counter.get_value()))

            if event == "return":
                track_value(arg, event)

        elif (is_ipython_frame(frame) or frame.f_back.f_code.co_filename in addl_filenames) and \
                event == "return":
            track_value(arg, event)

        if event == "return" and name in vars_not_found:
            varnames = vars_not_found.pop(name)
            for t, event_name, step in varnames:
                if t in frame.f_locals:
                    val = frame.f_locals[t]
                    track_value(val, event_name, step)

                elif t in frame.f_globals:
                    val = frame.f_globals[t]
                    track_value(val, event_name, step)

        return collect_intermidiate_results

    ACTIVE_FOOTPRINT = footprint
    return footprint, collect_intermidiate_results


def get_tracing_frame():
    """
    Returns the frame that is being traced by looking for the ``__PYBRYT_TRACING__`` global variable.

    Returns:
        the frame being traced or ``None`` of no tracing is occurring
    """
    frame = inspect.currentframe()
    while frame is not None:
        if TRACING_VARNAME in frame.f_globals and frame.f_globals[TRACING_VARNAME]:
            return frame
        frame = frame.f_back
    return None


def tracing_off(frame=None, save_func=True):
    """
    Turns off PyBryt's tracing if tracing is occurring in this call stack. If PyBryt is not tracing,
    takes no action.

    This method can be used in students' notebooks to include code that shouldn't be traced as part
    of the submission, e.g. demo code or ungraded code. In the example below, the call that creates
    ``x2`` is traced but the one to create ``x3`` is not.

    .. code-block:: python

        def pow(x, a):
            return x ** a

        x2 = pow(x, 2)

        pybryt.tracing_off()
        x3 = pow(x, 3)
    """
    global TRACING_FUNC
    frame = get_tracing_frame() if frame is None else frame
    if frame is None:
        return
    if save_func:
        TRACING_FUNC = frame.f_trace
    vn = f"sys_{make_secret()}"
    exec(f"import sys as {vn}\n{vn}.settrace(None)", frame.f_globals, frame.f_locals)


def tracing_on(frame=None, tracing_func=None):
    """
    Turns tracing on if PyBryt was tracing the call stack. If PyBryt is not tracing or
    :py:meth:`tracing_off<pybryt.tracing_off>` has not been called, no action is taken.

    This method can be used in students' notebooks to turn tracing back on after deactivating tracing
    for ungraded code In the example below, ``x4`` is traced because ``tracing_on`` is used after
    ``tracing_off`` and the creation of ``x3``.

    .. code-block:: python

        def pow(x, a):
            return x ** a

        x2 = pow(x, 2)

        pybryt.tracing_off()
        x3 = pow(x, 3)
        pybryt.tracing_on()

        x4 = pow(x, 4)
    """
    global TRACING_FUNC
    frame = get_tracing_frame() if frame is None else frame
    if frame is None or (TRACING_FUNC is None and tracing_func is None):
        return
    if TRACING_FUNC is not None and tracing_func is None:
        tracing_func = TRACING_FUNC
    vn = f"cir_{make_secret()}"
    vn2 = f"sys_{make_secret()}"
    frame.f_globals[vn] = tracing_func
    exec(f"import sys as {vn2}\n{vn2}.settrace({vn})", frame.f_globals, frame.f_locals)
    frame.f_trace = tracing_func


class no_tracing:
    """
    A context manager for turning tracing off for a block of code in a submission.

    If PyBryt is tracing code, any code inside this context will not be traced for values in memory.
    If PyBryt is not tracing, no action is taken.

    .. code-block:: python

        with pybryt.no_tracing():
            # this code is not traced
            foo(1)
        
        # this code is traced
        foo(2)
    """

    def __enter__(self):
        tracing_off()

    def __exit__(self, exc_type, exc_value, traceback):
        tracing_on()
        return False


class FrameTracer:
    """
    A class for managing the tracing of a call stack.

    Args:
        frame (``FrameType``): the frame to initialize tracing in
    """

    footprint: Optional[MemoryFootprint]
    """the memory footprint being populated"""

    frame: FrameType
    """the frame being traced"""

    _tracing_already_enabled: bool
    """whether tracing was already enabled when ``start_trace`` was called"""

    def __init__(self, frame: FrameType) -> None:
        self.frame = frame
        self.footprint = None
        self._tracing_already_enabled = False

    def start_trace(self, **kwargs) -> bool:
        """
        Create a collector and memory footprint and start tracing execution in the frame. Returns
        a boolean indicating whether tracing was enabled.

        Args:
            **kwargs: additional keyword arguments passed to ``create_collector``

        Returns:
            ``bool``: whether this call initiated tracing (``False`` if tracing was already enabled)
        """
        self._tracing_already_enabled = get_tracing_frame() is not None
        if self._tracing_already_enabled:
            self.footprint = get_active_footprint()
            return False

        self.footprint, cir = create_collector(**kwargs)
        self.frame.f_globals[TRACING_VARNAME] = True
        tracing_on(tracing_func=cir)
        return True

    def end_trace(self) -> None:
        """
        End execution tracing in the frame.
        """
        if not self._tracing_already_enabled:
            tracing_off(save_func=False)
            self.frame.f_globals[TRACING_VARNAME] = False

    def get_footprint(self) -> MemoryFootprint:
        """
        Return the memory footprint that was populated by the trace function.

        Returns:
            :py:class:`pybryt.execution.memory_footprint.MemoryFootprint`: the memory footprint
        """
        return self.footprint


def get_active_footprint() -> Optional[MemoryFootprint]:
    """
    Get the active memory footprint if present, else ``None``.

    Returns:
        :py:class:`pybryt.execution.memory_footprint.MemoryFootprint`: the memory footprint
    """
    return ACTIVE_FOOTPRINT


def set_initial_conditions(conditions: Dict[str, Any]) -> None:
    """
    Update the initial conditions stored in the active memory footprint if it exists.

    Args:
        conditions (``dict[str, object]``): a dictionary of initial conditions
    """
    footprint = get_active_footprint()
    if footprint is not None:
        footprint.set_initial_conditions(conditions)



"""Student implementations for PyBryt"""

__all__ = ["StudentImplementation", "check", "generate_student_impls"]

import hashlib
import inspect
import nbformat
import os
import warnings

from glob import glob
from multiprocessing import Process, Queue
from typing import Any, Dict, List, Optional, Union

from .execution import execute_notebook, FrameTracer, MemoryFootprint, NBFORMAT_VERSION
from .reference import generate_report, ReferenceImplementation, ReferenceResult
from .utils import Serializable


CACHE_DIR_NAME = ".pybryt_cache"
CACHE_STUDENT_IMPL_PREFIX = "student_impl_{}.pkl"


class StudentImplementation(Serializable):
    """
    A student implementation class for handling the execution of student work and manging the 
    memory footprint generated by that execution.

    Args:
        path_or_nb (``str`` or ``nbformat.NotebookNode``): the submission notebook or the path to it
        addl_filenames (``list[str]``, optional): additional filenames to trace inside during 
            execution
        output (``str``, optional): a path at which to write executed notebook
        timeout (``int``, optional): number of seconds to allow for notebook execution; set to 
            ``None`` for no time limit
    """

    nb: Optional[nbformat.NotebookNode]
    """the submission notebook"""

    nb_path: Optional[str]
    """the path to the notebook file"""

    footprint: MemoryFootprint
    """the memory footprint"""

    def __init__(
        self,
        path_or_nb: Optional[Union[str, nbformat.NotebookNode]],
        addl_filenames: List[str] = [],
        output: Optional[str] = None,
        timeout: Optional[int] = 1200,
    ):
        if path_or_nb is None:
            self.nb = None
            self.nb_path = None
            return
        if isinstance(path_or_nb, str):
            self.nb = nbformat.read(path_or_nb, as_version=NBFORMAT_VERSION)
            self.nb_path = path_or_nb
        elif isinstance(path_or_nb, nbformat.NotebookNode):
            self.nb = path_or_nb
            self.nb_path = ""
        else:
            raise TypeError(f"path_or_nb is of unsupported type {type(path_or_nb)}")

        self._execute(timeout, addl_filenames=addl_filenames, output=output)

    def _execute(
        self, 
        timeout: Optional[int], 
        addl_filenames: List[str] = [], 
        output: Optional[str] = None
    ) -> None:
        """
        Executes the notebook ``self.nb``.

        Args:
            timeout (``int``): number of seconds to allow for notebook execution; set to 
                ``None`` for no time limit
            addl_filenames (``list[str]``, optional): additional filenames to trace inside during 
                execution
            output (``str``, optional): a path at which to write executed notebook
        """
        self.footprint = execute_notebook(
            self.nb, 
            self.nb_path, 
            addl_filenames=addl_filenames, 
            timeout=timeout,
        )

        if output:
            with open(output, "w+") as f:
                nbformat.write(self.footprint.executed_notebook, f)

        if self.errors:
            nb_path = self.nb_path
            if not nb_path:
                nb_path = "student notebook"
            warnings.warn(f"Executing {nb_path} produced errors in the notebook")

    @property
    def errors(self) -> List[Dict[str, Union[str, List[str]]]]:
        """
        ``list[dict[str, Union[str, list[str]]]]:``: a list of error outputs from the executed notebook
        if present
        """
        if self.footprint.executed_notebook is None:
            return []

        errors = []
        for cell in self.footprint.executed_notebook['cells']:
            if cell['cell_type'] == "code":
                for out in cell['outputs']:
                    if out['output_type'] == "error":
                        errors.append(out)

        return errors

    @classmethod
    def from_footprint(cls, footprint: MemoryFootprint) -> 'StudentImplementation':
        """
        Create a student implementation object from a memory footprint directly, rather than by
        executing a notebook. Leaves the ``nb`` and ``nb_path`` instance variables of the resulting 
        object empty.

        Args:
            footprint (:py:class:`pybryt.execution.memory_footprint.MemoryFootprint`): the memory 
                footprint
        """
        stu = cls(None)
        stu.footprint = footprint
        return stu

    @classmethod
    def combine(cls, impls: List['StudentImplementation']) -> 'StudentImplementation':
        """
        Combine a series of student implementations into a single student implementation.

        Collects the memory footprint of each implementation into a single list and offsets the
        timestamp of each object by the total number of steps for each preceding student
        implementation in the list. 

        Assumes that the student implementations are provided in sorted order. Filters out duplicate
        values present in multiple implementations by keeping the earliest.

        Args:
            impls (``list[StudentImplementation]``): the list of implementations to combine

        Returns:
            ``StudentImplementation``: the combined implementation
        """
        footprint = MemoryFootprint.combine(*[impl.footprint for impl in impls])
        return cls.from_footprint(footprint)

    @classmethod
    def from_cache(cls, cache_dir=CACHE_DIR_NAME, combine=True) -> \
            Union['StudentImplementation', List['StudentImplementation']]:
        """
        Load one or more student implementations from a cache.

        All files are combined into a single student implementation by default, but a list can be
        returned instead by setting ``combine=False``.

        Args:
            cache_dir (``str``, optional): the path to the cache directory
            combine (``bool``, optional): whether to combine the implementations

        Returns:
            ``StudentImplementation`` or ``list[StudentImplementation]``: the student implementations
            loaded from the cache.
        """
        impls = []
        for impl_path in glob(os.path.join(cache_dir, CACHE_STUDENT_IMPL_PREFIX.format("*"))):
            impls.append(cls.load(impl_path))
        if combine:
            return cls.combine(impls)
        else:
            return impls

    def __eq__(self, other: Any) -> bool:
        """
        Checks whether a student implementation is equal to another object.

        The object is considered equal if it is also a student implementation, has the same memory
        footprint, the same number of steps, and the same source notebook.
        """
        return isinstance(other, type(self)) and self.footprint == other.footprint \
            and self.nb == other.nb

    @property
    def _default_dump_dest(self) -> str:
        return "student.pkl"

    def check(self, ref: Union[ReferenceImplementation, List[ReferenceImplementation]], group=None) -> \
            Union[ReferenceResult, List[ReferenceResult]]:
        """
        Checks this student implementation against a single or list of reference implementations.
        Returns the :py:class:`ReferenceResult<pybryt.ReferenceResult>` object(s) resulting from 
        those checks.

        Args:
            ref (``ReferenceImplementation`` or ``list[ReferenceImplementation]``): the reference(s)
                to run against
            group (``str``, optional): if specified, only annotations in this group will be run

        Returns:
            ``ReferenceResult`` or ``list[ReferenceResult]``: the results of the reference 
            implementation checks
        """
        if isinstance(ref, ReferenceImplementation):
            return ref.run(self.footprint, group=group)
        elif isinstance(ref, list):
            return [r.run(self.footprint, group=group) for r in ref]
        else:
            raise TypeError(f"check cannot take values of type {type(ref)}")

    def check_plagiarism(self, student_impls: List["StudentImplementation"], **kwargs) -> List[ReferenceResult]:
        """
        Checks this student implementation against a list of other student implementations for 
        plagiarism. Uses :py:meth:`create_references<pybryt.plagiarism.create_references>` to create
        a randomly-generated reference implementation from this student implementation and runs it
        against each of the implementations in ``student_impls`` using 
        :py:meth:`get_impl_results<pybryt.plagiarism.get_impl_results>`.

        Args:
            student_impls (``list[StudentImplementation]``): other student implementations to run
                against
            **kwargs: keyword arguments passed to 
                :py:meth:`create_references<pybryt.plagiarism.create_references>` and 
                :py:meth:`get_impl_results<pybryt.plagiarism.get_impl_results>`
        
        Returns:
            ``list[ReferenceResult]`` or ``numpy.ndarray``: the results of each student 
            implementation in ``student_impls`` when run against this student implementation
        """
        refs = create_references([self], **kwargs)
        return get_impl_results(refs[0], student_impls, **kwargs)


def generate_student_impls(
    paths_or_nbs: List[Union[str, nbformat.NotebookNode]], parallel: bool = False, **kwargs
) -> List[StudentImplementation]:
    """
    Generates multiple student implementations from a list of file paths or notebooks.

    Can optionally generate the student implementations in parallel processes using Python's 
    ``multiprocessing`` library to reduce the runtime.

    Args:
        paths_or_nbs (``list[Union[str, nbformat.NotebookNode]]``): the notebooks or paths to them
        parallel (``bool``, optional): whether to execute in parallel
        **kwargs: additional keyword arguments passed to the 
            :py:class:`StudentImplementation<pybryt.StudentImplementation>` constructor

    Returns:
        ``list[StudentImplementation]``: the student implementations
    """
    if parallel:
        def create_and_collect_impl(impl, queue, **kwargs):
            stu = StudentImplementation(impl, **kwargs)
            queue.put((impl, stu))

    impls = []
    if parallel:
        queue = Queue()

    for stu in paths_or_nbs:
        if parallel:
            p = Process(target=create_and_collect_impl, args=(stu, queue), kwargs=kwargs)
            p.start()
            impls.append(p)
        else:
            impls.append(StudentImplementation(stu, **kwargs))

    if parallel:
        for p in impls:
            p.join()

        procs, impls = impls, {}
        while not queue.empty():
            t = queue.get()
            impls[str(t[0])] = t[1]

        impls = [impls[str(p)] for p in paths_or_nbs]

    return impls


class check:
    """
    A context manager for running a block of student code against a reference implementation.

    This context manager, designed to be used in students' notebooks, can be used to check a block
    of student code against one or a series of reference implementations. The context manager turns
    on tracing before launching the block and captures the memory footprint that is created while
    the block is executed. It prints out a report upon exiting with information about the passing or
    failing of each reference and the messages returned.

    As an example, the block below uses this context manager to test a student's implementation of
    a Fibonacci generator ``fiberator``:

    .. code-block:: python

        with pybryt.check("fiberator.pkl"):
            fib = fiberator()
            for _ in range(100):
                next(fib)

    Args:
        ref (``Union[str, ReferenceImplementation, list[str], list[ReferenceImplementation]]``): the
            reference(s) to check against or the path(s) to them
        group (``str``, optional): the name of the group of annotations to run in each reference
        report_on_error (``bool``, optional): whether to print the report when an error is thrown
            by the block
        show_only (one of ``{'satisfied', 'unsatisfied', None}``, optional): which types of
            reference results to include in the report; if ``None``, all are included
        **kwargs: additional keyword arguments passed to ``pybryt.execution.create_collector``
    """

    _ref: List[ReferenceImplementation]
    """the references being checked against"""

    _group: Optional[str]
    """the group of annotations in the references to run"""

    _report_on_error: bool
    """whether to print the report when an error is thrown by the block"""

    _show_only: Optional[str]
    """which types of reference results to include in the report"""

    _frame_tracer: Optional[FrameTracer]
    """the frame tracer being used to manage tracing"""

    _cache: bool
    """whether to cache the memory footprint and results"""

    _kwargs: Dict[str, Any]
    """keyword arguments passed to ``pybryt.execution.create_collector``"""

    _disabled: bool
    """whether this check is disbaled (because PyBryt is already tracing)"""

    def __init__(
        self,
        ref: Union[str, ReferenceImplementation, List[str], List[ReferenceImplementation]], 
        group: Optional[str] = None,
        report_on_error: bool = True,
        show_only: Optional[str] = None,
        cache: bool = True,
        **kwargs,
    ):
        if isinstance(ref, str):
            ref = ReferenceImplementation.load(ref)
        if isinstance(ref, ReferenceImplementation):
            ref = [ref]
        if isinstance(ref, list):
            if len(ref) == 0:
                raise ValueError("Cannot check against an empty list of references")
            if not all(isinstance(r, ReferenceImplementation) for r in ref):
                if not all(isinstance(r, str) for r in ref):
                    raise TypeError("Invalid values in the reference list")
                ref = [ReferenceImplementation.load(r) for r in ref]
        if not all(isinstance(r, ReferenceImplementation) for r in ref):  # pragma: no cover
            raise TypeError("Invalid values provided for reference(s)")

        self._ref = ref
        self._group = group
        self._kwargs = kwargs
        self._show_only = show_only
        self._report_on_error = report_on_error
        self._cache = cache
        self._disabled = False

    def _cache_check(self, stu, res):
        """
        Cache the student implementation and reference results in the PyBryt cache directory.

        Args:
            stu (``StudentImplementation``): the student implementation created by the check
            res (``Union[ReferenceImplementation, list[ReferenceImplementation]]``): the reference 
                results
        """
        os.makedirs(CACHE_DIR_NAME, exist_ok=True)  # create cache directory if needed

        if not isinstance(res, list):
            res = [res]

        for r in res:
            res_path = os.path.join(CACHE_DIR_NAME, f"{r.name}_results.pkl")
            r.dump(res_path)

        ref_hash = hashlib.sha1("".join(r.name for r in res).encode()).hexdigest()
        stu_path = os.path.join(CACHE_DIR_NAME, CACHE_STUDENT_IMPL_PREFIX.format(ref_hash))
        stu.dump(stu_path)

    def __enter__(self):
        self._frame_tracer = FrameTracer(inspect.currentframe().f_back)
        self._disabled = not self._frame_tracer.start_trace(**self._kwargs)

    def __exit__(self, exc_type, exc_value, traceback):
        if not self._disabled:
            self._frame_tracer.end_trace()

            if exc_type is None or self._report_on_error:
                footprint = self._frame_tracer.get_footprint()
                stu = StudentImplementation.from_footprint(footprint)
                res = stu.check(self._ref, group=self._group)
                report = generate_report(res, show_only=self._show_only)
                if report:
                    print(report)

                if self._cache:
                    self._cache_check(stu, res)

        return False


from .plagiarism import create_references, get_impl_results



"""Tests for the imports preprocessor"""

import nbformat

from textwrap import dedent

from pybryt.preprocessors import ImportFindingPreprocessor


def test_preprocessor():
    """
    """
    nb = nbformat.v4.new_notebook()
    nb.cells.append(nbformat.v4.new_code_cell(dedent("""\
        import numpy
        import pandas as pd
        from random import choice
        from itertools import chain as c
        import statsmodels.api
        exec("import tqdm")
    """)))

    ifp = ImportFindingPreprocessor()

    nb2 = ifp.preprocess(nb)
    assert nb2 is nb
    assert ifp.imports == {"numpy", "pandas", "random", "itertools", "statsmodels"}



"""Tests for debug mode"""

from pybryt import debug_mode, disable_debug_mode, enable_debug_mode
from pybryt.debug import _debug_mode_enabled


def test_debug_mode():
    """
    Test that the debug mode controllers correctly enable and disable debug mode.
    """
    enable_debug_mode()
    assert _debug_mode_enabled()
    disable_debug_mode()
    assert not _debug_mode_enabled()

    with debug_mode():
        assert _debug_mode_enabled()
    assert not _debug_mode_enabled()

    enable_debug_mode()
    with debug_mode():
        assert _debug_mode_enabled()
    assert not _debug_mode_enabled()



""""""

import pytest
import random
import tempfile

from textwrap import dedent
from unittest import mock

from pybryt.utils import *

from .test_reference import generate_reference_notebook


def test_filter_pickleable_list():
    """
    """
    l = [1, 2, 3]
    filter_pickleable_list(l)
    assert len(l) == 3

    with mock.patch("dill.dumps") as mocked_dill:
        mocked_dill.side_effect = Exception()
        filter_pickleable_list(l)
        assert len(l) == 0


def test_notebook_to_string():
    """
    """
    ref = generate_reference_notebook()
    s = notebook_to_string(ref)
    assert s.strip() == dedent("""\
        import pybryt

        def median(S):
            sorted_S = sorted(S) 
            pybryt.Value(sorted_S, name="sorted", group="median", limit=5, success_message="SUCCESS: Sorted the sample correctly", 
                        failure_message="ERROR: The sample was not sorted")

            size_of_set = len(S) 
            pybryt.Value(size_of_set, name="size", group="median", success_message = "SUCCESS: Computed the size of the sample", 
                        failure_message="ERROR: Did not capture the size of the set to determine if it is odd or even")

            middle = size_of_set // 2
            is_set_size_even = (size_of_set % 2) == 0

            if is_set_size_even:
                return (sorted_S[middle-1] + sorted_S[middle]) / 2
            else:
                return sorted_S[middle]

        import numpy as np
        np.random.seed(42)
        for _ in range(10):
            vals = [np.random.randint(-1000, 1000) for _ in range(np.random.randint(1, 1000))]
            val = median(vals)
            pybryt.Value(val, name="median", group="median", success_message="SUCCESS: computed the correct median", 
                failure_message="ERROR: failed to compute the median")
    """).strip()

    with pytest.raises(TypeError, match="invalid notebook type"):
        notebook_to_string(1)


def test_make_secret():
    """
    """
    random.seed(42)
    s = make_secret()
    assert s == "HBRPOI"


def test_save_notebook():
    """
    """
    with mock.patch("pybryt.utils.get_ipython") as mocked_get:
        with mock.patch("pybryt.utils.publish_display_data") as mocked_pub:
            mocked_get.return_value = True
            with tempfile.NamedTemporaryFile(suffix=".ipynb") as ntf:
                v = save_notebook(ntf.name, timeout=1)
                mocked_pub.assert_called()
                assert not v



""""""

import numpy as np
import sys

from unittest import mock

from pybryt import MemoryFootprint, no_tracing, set_initial_conditions
from pybryt.execution import create_collector, FrameTracer, tracing_off, tracing_on
from pybryt.execution.memory_footprint import Event

from .utils import generate_mocked_frame


__PYBRYT_TRACING__ = False


def test_trace_function():
    """
    """
    np.random.seed(42)
    
    tracked_filepath = "/path/to/tracked/file.py"

    frame = generate_mocked_frame("<ipython-abc123>", "foo", 3)
    footprint, cir = create_collector(addl_filenames=[tracked_filepath])

    arr = np.random.uniform(-100, 100, size=(100, 100))
    cir(frame, "return", arr)
    assert len(footprint) == 1
    assert np.allclose(footprint.get_value(0).value, arr)
    assert footprint.get_value(0).timestamp == 1
    assert footprint.get_value(0).event == Event.RETURN

    # test value in skip_types
    cir(frame, "return", type(1))
    assert len(footprint) == 1

    # test pickling error
    with mock.patch("dill.dumps") as mocked_dumps:
        mocked_dumps.side_effect = Exception()
        cir(frame, "return", 1)
    
    assert len(footprint) == 1

    # test call event
    assert len(footprint.calls) == 0
    cir(frame, "call", None)
    assert len(footprint.calls) == 1
    assert footprint.calls[0] == ("<ipython-abc123>", "foo")

    frame = generate_mocked_frame(
        "<ipython-abc123>", "foo", 3, {"data": arr}
    )

    # check line processing working
    with mock.patch("linecache.getline") as mocked_linecache:

        # check eval call for attributes
        mocked_linecache.return_value = "data.T"
        cir(frame, "line", None)
        assert len(footprint) == 2
        assert np.allclose(footprint.get_value(1).value, arr.T)
        assert footprint.get_value(1).timestamp == 5
        assert footprint.get_value(1).event == Event.LINE

        # check failed eval call for attributes
        mocked_linecache.return_value = "data.doesnt_exist"
        cir(frame, "line", None)
        assert len(footprint) == 2

        # check looking in frame locals + globals
        frame.f_globals["more_data"] = np.random.uniform(-100, 100, size=(100, 100))
        frame.f_locals["more_data"] = np.random.uniform(-100, 100, size=(100, 100))
        mocked_linecache.return_value = "more_data"
        cir(frame, "line", None)
        assert len(footprint) == 3
        assert np.allclose(footprint.get_value(2).value, frame.f_locals["more_data"])
        assert footprint.get_value(2).timestamp == 7

        # check that we track assignment statements on function return
        mocked_linecache.return_value = "even_more_data = more_data ** 2"
        cir(frame, "line", None)
        assert len(footprint) == 3

        mocked_linecache.return_value = "even_more_data_2 = more_data ** 3"
        cir(frame, "line", None)
        assert len(footprint) == 3

        # check that floats aren't added with the eval call
        mocked_linecache.return_value = "event_more_data_3 = [2.1, 1000, 100.3]"
        cir(frame, "line", None)
        assert len(footprint) == 3

        frame.f_locals["even_more_data"] = frame.f_locals["more_data"] ** 2
        frame.f_globals["even_more_data_2"] = frame.f_locals["more_data"] ** 3
        mocked_linecache.return_value = ""
        cir(frame, "return", None)
        assert len(footprint) == 6
        assert footprint.get_value(3).value is None
        assert footprint.get_value(3).timestamp == 11
        assert np.allclose(footprint.get_value(4).value, frame.f_locals["more_data"] ** 2)
        assert footprint.get_value(4).timestamp == 8
        assert np.allclose(footprint.get_value(5).value, frame.f_locals["more_data"] ** 3)
        assert footprint.get_value(5).timestamp == 9

        # check that skip_types respected
        frame.f_locals["none_type"] = type(None)
        mocked_linecache.return_vaue = "none_type"
        cir(frame, "line", None)
        assert len(footprint) == 6

        # check that addl_filenames respected
        frame = generate_mocked_frame(tracked_filepath, "bar", 100, f_back=frame)
        frame.f_locals["data"] = arr
        mocked_linecache.return_value = "arr = -1 * data"
        cir(frame, "line", None)
        frame.f_locals["arr"] = -1 * arr
        cir(frame, "return", None) # run a return since arr shows up in vars_not_found
        assert len(footprint) == 7
        assert np.allclose(footprint.get_value(6).value, -1 * arr)
        assert footprint.get_value(6).timestamp == 14
    
    # check that IPython child frame return values are tracked
    frame = generate_mocked_frame("/path/to/file.py", "bar", 100, f_back=frame)
    cir(frame, "line", None)
    assert len(footprint) == 7

    cir(frame, "return", np.exp(arr))
    assert len(footprint) == 8
    assert np.allclose(footprint.get_value(7).value, np.exp(arr))
    assert footprint.get_value(7).timestamp == 14

    assert len(footprint.calls) == 1
    frame = generate_mocked_frame("/path/to/foo.py", "bar", 100)
    cir(frame, "call", None)
    assert len(footprint.calls) == 2
    assert footprint.calls[0] == ("<ipython-abc123>", "foo")
    assert footprint.calls[1] == ("/path/to/foo.py", "bar")

    # test ipykernel v6 frame filename
    frame = generate_mocked_frame("/var/8k/ipykernel_495995/29304985.py", "foo", 3)
    footprint, cir = create_collector(addl_filenames=[tracked_filepath])

    arr = np.random.uniform(-100, 100, size=(100, 100))
    cir(frame, "return", arr)
    assert len(footprint) == 1
    assert np.allclose(footprint.get_value(0).value, arr)
    assert footprint.get_value(0).timestamp == 1

    # check that imported modules are correctly tracked
    with mock.patch.object(footprint, "add_imports") as mocked_add_imports:
        class Foo:
            pass

        cir(frame, "return", Foo)
        mocked_add_imports.assert_called_with(Foo.__module__.split(".")[0])

        cir(frame, "return", np)
        mocked_add_imports.assert_called_with(np.__name__)


def test_tracing_control():
    """
    """
    global __PYBRYT_TRACING__
    trace = lambda frame, event, arg: trace

    __PYBRYT_TRACING__ = True
    with mock.patch("sys.settrace") as mocked_settrace:
        tracing_off()
        mocked_settrace.assert_called()

    with mock.patch("sys.settrace") as mocked_settrace:
        tracing_on()
        if sys.gettrace() is not None:
            mocked_settrace.assert_called()

    __PYBRYT_TRACING__ = False
    with mock.patch("sys.settrace") as mocked_settrace:
        tracing_off()
        mocked_settrace.assert_not_called()

    with mock.patch("sys.settrace") as mocked_settrace:
        tracing_on()
        mocked_settrace.assert_not_called()


def test_tracing_context_manager():
    """
    """
    with mock.patch("pybryt.execution.tracing.tracing_off") as mocked_off, \
            mock.patch("pybryt.execution.tracing.tracing_on") as mocked_on:
        with no_tracing():
            mocked_off.assert_called()
            mocked_on.assert_not_called()
        mocked_on.assert_called()


def test_frame_tracer():
    """
    Tests for ``pybryt.execution.tracing.FrameTracer``.
    """
    frame = generate_mocked_frame("<ipython-abc123>", "foo", 3)
    tracer = FrameTracer(frame)

    with mock.patch("pybryt.execution.tracing.tracing_on") as mocked_on, \
            mock.patch("pybryt.execution.tracing.tracing_off") as mocked_off, \
            mock.patch("pybryt.execution.tracing.create_collector") as mocked_create, \
            mock.patch("pybryt.execution.tracing.get_tracing_frame") as mocked_get:
        kwargs = {"foo": "bar"}
        mocked_create.return_value = (MemoryFootprint(), lambda a, b, c: None)

        # check that when tracing is enabled no action is taken
        mocked_get.return_value = frame
        assert not tracer.start_trace()
        mocked_on.assert_not_called()

        tracer.end_trace()
        mocked_off.assert_not_called()

        mocked_get.return_value = None

        assert tracer.start_trace(**kwargs)
        mocked_create.assert_called_once_with(**kwargs)
        mocked_on.assert_called_once_with(tracing_func=mocked_create.return_value[1])
        assert frame.f_globals["__PYBRYT_TRACING__"]
        assert isinstance(tracer.footprint, MemoryFootprint)

        tracer.end_trace()
        mocked_off.assert_called_once_with(save_func=False)
        assert not frame.f_globals["__PYBRYT_TRACING__"]

        assert isinstance(tracer.get_footprint(), MemoryFootprint)


def test_set_initial_conditions():
    """
    Tests for ``pybryt.execution.tracing.set_initial_conditions``.
    """
    with mock.patch("pybryt.execution.tracing.get_active_footprint") as mocked_get:
        ics = {"foo": 1, "bar": 2}
        set_initial_conditions(ics)
        mocked_get.return_value.set_initial_conditions.assert_called_with(ics)

        mocked_get.return_value = None
        set_initial_conditions(ics)



"""Tests for PyBryt execution internals"""

import dill
import nbformat
import numpy as np
import pathlib
import random
import tempfile

from unittest import mock

import pybryt.execution


def generate_test_notebook():
    """
    """
    nb = nbformat.v4.new_notebook()
    nb.cells.append(nbformat.v4.new_code_cell(
        "import numpy as np\nimport pandas as pd\nimport matplotlib.pyplot as plt\n%matplotlib inline"
        ))
    nb.cells.append(nbformat.v4.new_code_cell(
        "np.random.seed(42)\nx = np.random.uniform(size=1000)\ny = np.random.normal(size=1000)"
    ))
    nb.cells.append(nbformat.v4.new_code_cell("df = pd.DataFrame({'x': x, 'y': y})"))
    return nb


def test_notebook_execution():
    """
    """
    random.seed(42)
    nb = generate_test_notebook()

    with tempfile.NamedTemporaryFile(delete=False) as observed_ntf:
        with mock.patch("pybryt.execution.mkstemp") as mocked_tempfile:
            mocked_tempfile.return_value = (None, observed_ntf.name)

            footprint = pybryt.execution.execute_notebook(nb, "")
            assert len(footprint) > 0
            assert all(i in footprint.imports for i in ["pandas", "numpy", "matplotlib"])
            assert len(footprint.calls) > 0



"""Execution utilities"""

import os

from types import FrameType


def is_ipython_frame(frame: FrameType) -> bool:
    """
    Determine whether a frame is being executed by IPython.

    Args:
        frame (``types.FrameType``): the frame to examine

    Returns:
        ``bool``: whether the frame is an IPython frame
    """
    filename = frame.f_code.co_filename
    parent_dir = os.path.split(os.path.split(filename)[0])[1]
    return filename.startswith("<ipython") or parent_dir.startswith("ipykernel_")



"""Tests for value annotations"""

import pytest

from itertools import chain
from unittest import mock

from pybryt import Annotation, Attribute, debug_mode, InitialCondition, ReturnValue, Value
from pybryt import invariants as inv
from pybryt.execution.memory_footprint import Event, MemoryFootprint, MemoryFootprintValue
from pybryt.utils import pickle_and_hash

from .utils import assert_object_attrs, generate_memory_footprint


def test_value_annotation():
    """
    """
    # TODO: create pytest fixtures that do these automatically
    footprint = generate_memory_footprint()
    Annotation.reset_tracked_annotations()

    seen = {}
    for mfp_val in footprint:
        v = Value(mfp_val.value)
        res = v.check(footprint)

        assert repr(res) == "AnnotationResult(satisfied=True, annotation=pybryt.Value)"

        h = pickle_and_hash(mfp_val.value)

        # check attributes of BeforeAnnotation and AnnotationResult
        assert_object_attrs(v, {"children__len": 0})
        assert_object_attrs(res, {
            "children": [],
            "satisfied": True,
            "_satisfied": True,
            "annotation": v,
            "timestamp": seen[h] if h in seen else mfp_val.timestamp,
            "value": mfp_val.value,
        })

        if h not in seen:
            seen[h] = mfp_val.timestamp

    v = Value(-1)  # does not occur in footprint
    res = v.check(footprint)

    assert v.to_dict() == {
        "name": "Annotation 11",
        "children": [],
        "success_message": None,
        "failure_message": None,
        "limit": None,
        "group": None,
        "invariants": [],
        "atol": None,
        "rtol": None,
        "type": "value",
    }
    assert repr(res) == "AnnotationResult(satisfied=False, annotation=pybryt.Value)"

    # check __repr__
    assert repr(v) == "pybryt.Value", "wrong __repr__"

    # check attributes of BeforeAnnotation and AnnotationResult
    assert_object_attrs(v, {"children__len": 0})
    assert_object_attrs(res, {
        "children": [],
        "satisfied": False,
        "_satisfied": False,
        "annotation": v,
        "timestamp": -1,
        "value": None,
    })

    # test pickling error
    with mock.patch("dill.dumps") as mocked_dumps:
        mocked_dumps.side_effect = Exception()
        with pytest.raises(ValueError):
            v = Value(-1)

    # test with invariants
    s = footprint.get_value(-1).value
    v = Value(s.upper(), invariants=[inv.string_capitalization])
    res = v.check(footprint)
    assert res.satisfied

    # test that check_against correctly calls check
    with mock.patch.object(v, "check") as mocked_check:
        mocked_check.return_value = mock.MagicMock()
        mocked_check.return_value.satisfied = True
        assert v.check_against(s.lower())
        mocked_check.assert_called_with(MemoryFootprint.from_values(MemoryFootprintValue(s.lower(), 0, None)))

    # check custom equivalence function
    mocked_eq = mock.MagicMock()
    v = Value(s, equivalence_fn=mocked_eq)
    mocked_eq.return_value = False
    assert not v.check_against("foo")
    mocked_eq.assert_called_with(s, "foo")
    mocked_eq.return_value = True
    assert v.check_against("")
    mocked_eq.assert_called_with(s, "")
    mocked_eq.side_effect = ValueError()
    assert not v.check_against("")

    # check for invalid return type error
    mocked_eq.return_value = 1
    mocked_eq.side_effect = None
    with pytest.raises(TypeError, match=f"Custom equivalence function returned value of invalid type: {type(1)}"):
        v.check_against(1)

    # check debug mode errors
    with debug_mode():
        with pytest.raises(ValueError, match="Absolute or relative tolerance specified with an equivalence function"):
            Value(1, atol=1e-5, equivalence_fn=lambda x, y: True)

        with pytest.raises(ValueError, match="Absolute or relative tolerance specified with an equivalence function"):
            Value(1, rtol=1e-5, equivalence_fn=lambda x, y: True)

        class FooError(Exception):
            pass

        with pytest.raises(FooError):
            def raise_foo(x, y):
                raise FooError()

            v = Value(1, equivalence_fn=raise_foo)
            v.check_against(1)


def test_attribute_annotation():
    """
    """
    footprint = generate_memory_footprint()
    Annotation.reset_tracked_annotations()
    val, ts = footprint.get_value(0).value, footprint.get_value(0).timestamp

    v = Attribute(val, "T")
    res = v.check(footprint)

    # check attributes of BeforeAnnotation and AnnotationResult
    assert_object_attrs(v, {"children__len": 1})
    assert_object_attrs(res, {
        "children__len": 1,
        "satisfied": True,
        "_satisfied": None,
        "annotation": v,
        "timestamp": -1,
        "satisfied_at": ts,
        "value": val,
    })

    assert v.to_dict() == {
        "name": "Annotation 2",
        "children": [
            {
                'name': 
                'Annotation 1', 
                'group': None, 
                'limit': None, 
                'success_message': None, 
                'failure_message': None, 
                'children': [], 
                'invariants': [], 
                'atol': None,
                'rtol': None,
                "type": None,
            }
        ],
        "success_message": None,
        "failure_message": None,
        "limit": None,
        "group": None,
        "invariants": [],
        "atol": None,
        "rtol": None,
        "type": "attribute",
        "attributes": ['T'],
        "enforce_type": False,
    }

    # test that check_against correctly calls check
    with mock.patch.object(v, "check") as mocked_check:
        mocked_check.return_value = mock.MagicMock()
        mocked_check.return_value.satisfied = False
        assert not v.check_against(val)
        mocked_check.assert_called_with(MemoryFootprint.from_values(MemoryFootprintValue(val, 0, None)))

    # check enforce type
    class Foo:
        T = val.T

    footprint2 = MemoryFootprint.from_values(MemoryFootprintValue(Foo(), 1, None))
    res = v.check(footprint2)
    assert res.satisfied

    v = Attribute(val, "T", enforce_type=True)
    res = v.check(footprint2)
    assert not res.satisfied

    footprint3 = MemoryFootprint.from_values(*chain(footprint, footprint2))
    res = v.check(footprint3)
    assert res.satisfied

    # check error raising
    with pytest.raises(TypeError):
        Attribute(val, ["T", 1])
    
    with pytest.raises(TypeError):
        Attribute(val, 1)

    with pytest.raises(AttributeError):
        Attribute(val, "foo")


def test_return_value():
    """
    Tests for the ``ReturnValue`` annotation.
    """
    rv = ReturnValue(1)
    footprint = MemoryFootprint.from_values(
        MemoryFootprintValue(1, 0, Event.RETURN),
        MemoryFootprintValue(2, 1, Event.LINE_AND_RETURN),
    )

    res = rv.check(footprint)
    assert_object_attrs(res, {
        "children": [],
        "satisfied": True,
        "_satisfied": True,
        "annotation": rv,
        "timestamp": 0,
        "value": 1,
    })

    footprint = MemoryFootprint.from_values(
        MemoryFootprintValue(1, 0, Event.LINE_AND_RETURN),
        MemoryFootprintValue(2, 1, Event.RETURN),
    )

    res = rv.check(footprint)
    assert_object_attrs(res, {
        "children": [],
        "satisfied": True,
        "_satisfied": True,
        "annotation": rv,
        "timestamp": 0,
        "value": 1,
    })

    footprint = MemoryFootprint.from_values(
        MemoryFootprintValue(1, 0, Event.LINE),
        MemoryFootprintValue(2, 1, Event.RETURN),
    )

    res = rv.check(footprint)
    assert_object_attrs(res, {
        "children": [],
        "satisfied": False,
        "_satisfied": False,
        "annotation": rv,
        "timestamp": -1,
        "value": None,
    })


def test_initial_conditions():
    """
    Tests for the use of initial conditions with value annotations.
    """
    name = "foo"
    ic = InitialCondition(name)

    val, offset = 2, 3
    v = Value(ic + offset)

    fp = MemoryFootprint.from_values(MemoryFootprintValue(val + offset, 0, None))
    fp.set_initial_conditions({name: val})

    res = v.check(fp)
    assert res.satisfied
    assert res.satisfied_at == 0
    assert res.value == val + offset

    ic2 = InitialCondition(name)
    v1, v2 = Value(ic, name="baz"), Value(ic2, name="baz")
    assert v1 == v2
    
    ic2 = InitialCondition("bar")
    v2 = Value(ic2, name="baz")
    assert v1 != v2

    # test errors
    with pytest.raises(ValueError, match="check_against cannot be used with initial conditions"):
        v.check_against(val + offset)

    with pytest.raises(TypeError, match="Initial conditions are not compatible with attribute annotations"):
        Attribute(ic, "foo")



"""Tests for ``pybryt.complexity``"""

from unittest import mock

from pybryt import complexities as cplx
from pybryt.complexity import ANNOTATION_NAME, _check_time_complexity_wrapper, TimeComplexityChecker
from pybryt.execution import MemoryFootprintValue, TimeComplexityResult


def test_time_complexity_checker():
    """
    Tests for ``pybryt.complexity.TimeComplexityChecker``.
    """
    checker = TimeComplexityChecker()
    
    with mock.patch("pybryt.complexity._check_time_complexity_wrapper") as mocked_wrapper:
        n = 10
        checker(n)
        mocked_wrapper.assert_called_once_with(checker, n)

        res = TimeComplexityResult("foo", n, 0, 10)
        checker.add_result(res)
        assert checker.results == [res]

        with mock.patch("pybryt.complexity.TimeComplexity") as mocked_annot:
            ret = checker.determine_complexity()
            mocked_annot.assert_called_once_with(cplx.constant, name=ANNOTATION_NAME)
            mocked_annot.return_value.check.assert_called()
            assert ret == mocked_annot.return_value.check.return_value.value


def test_wrapper():
    """
    Tests for ``pybryt.complexity._check_time_complexity_wrapper``.
    """
    n = 10
    checker = TimeComplexityChecker()
    res = TimeComplexityResult(ANNOTATION_NAME, n, 0, 10)
    with mock.patch.object(checker, "add_result") as mocked_add, \
            mock.patch("pybryt.complexity.FrameTracer") as mocked_tracer, \
            mock.patch("pybryt.complexity.inspect") as mocked_inspect, \
            mock.patch("pybryt.complexity.check_time_complexity") as mocked_cm:
        mocked_tracer.return_value.get_footprint.return_value.__iter__ = \
            mock.Mock(return_value=iter([MemoryFootprintValue(res, 0, None)]))
        wrapper = _check_time_complexity_wrapper(checker, n)
        with wrapper:
            mocked_tracer.assert_called_once_with(mocked_inspect.currentframe.return_value.f_back)
            mocked_tracer.return_value.start_trace.assert_called_once()
            mocked_cm.assert_called_once_with(ANNOTATION_NAME, n)
            mocked_cm.return_value.__enter__.assert_called_once()

        mocked_cm.return_value.__exit__.assert_called_once()
        mocked_tracer.return_value.end_trace.assert_called_once()
        mocked_tracer.return_value.get_footprint.assert_called_once()
        mocked_add.assert_called_once_with(res)



"""Complexity classes for complexity annotations"""

import numpy as np

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, List, Union


@dataclass
class ComplexityClassResult:
    """
    A container class for the results of complexity checks.
    """

    complexity_class: "complexity"
    """the complexity class associated with this result"""

    residual: float
    """the residual from the least-squares fit"""


class complexity(ABC):
    """
    Abstract base class for a complexity. Subclassses should implement the ``transform_n`` method,
    which transforms the input lengths array so that least squares can be used. If needed, the
    ``transform_t`` method can also be overwritten to transform the step counter values.

    The architecture for these and the algorithm for determining the optimal complexity is borrowed
    from https://github.com/pberkes/big_O.
    """

    def __or__(self, other: Any) -> "ComplexityUnion":
        """
        Create a complexity union through use of the ``|`` operator.

        Args:
            other (any): the other object in the union

        Returns:
            :py:class:`ComplexityUnion`: the union
        """
        return ComplexityUnion.from_or(self, other)

    def __call__(self, complexity_data: Dict[int, int]) -> ComplexityClassResult:
        """
        Return the sum of residuals by performing least squares on the input length data and timings.

        Uses ``transform_n`` and ``transform_t`` to transform the data from ``complexity_data``, which
        is a dictionary mapping input lengths to step counts. Performs least squares on the resulting
        arrays and returns the sum of residuals.

        Args:
            complexity_data (``dict[int, int]``): the complexity information

        Returns:
            ``float``: the sum of residuals from least squares
        """
        ns, ts = [], []
        for n, t in complexity_data.items():
            ns.append(n)
            ts.append(t)
        
        n = np.array(ns, dtype=int)
        t = np.array(ts, dtype=int)

        try:
            n = self.transform_n(n)
            t = self.transform_t(t)

            _, resid, _, _ = np.linalg.lstsq(n, t, rcond=-1)
            if len(resid) == 0:
                return ComplexityClassResult(self, np.inf)
            return ComplexityClassResult(self, resid[0])
        except:
            return ComplexityClassResult(self, np.inf)

    @staticmethod
    @abstractmethod
    def transform_n(n: np.ndarray) -> np.ndarray:
        """
        Transforms the array of input lengths for performing least squares.

        Args:
            n (``np.ndarray``): the array of input lengths
        
        Returns:
            ``np.ndarray``: the transformed array of input lengths
        """
        ... # pragma: no cover

    @staticmethod
    def transform_t(t: np.ndarray) -> np.ndarray:
        """
        Transforms the array of timings for performing least squares.

        Args:
            n (``np.ndarray``): the array of timings
        
        Returns:
            ``np.ndarray``: the transformed array of timings
        """
        return t


class _constant(complexity):
    """
    Complexity class for constant time: :math:`\\mathcal{O}(1)`
    """

    @staticmethod
    def transform_n(n: np.ndarray) -> np.ndarray:
        return np.ones((len(n), 1))


# complexity classes are singletons
constant = _constant()


class _logarithmic(complexity):
    """
    Complexity class for logarithmic time: :math:`\\mathcal{O}(\\log n)`
    """

    @staticmethod
    def transform_n(n: np.ndarray) -> np.ndarray:
        return np.vstack((np.ones(len(n)), np.log2(n))).T


logarithmic = _logarithmic()


class _linear(complexity):
    """
    Complexity class for linear time: :math:`\\mathcal{O}(n)`
    """

    @staticmethod
    def transform_n(n: np.ndarray) -> np.ndarray:
        return np.vstack((np.ones(len(n)), n)).T


linear = _linear()


class _linearithmic(complexity):
    """
    Complexity class for linearithmic time: :math:`\\mathcal{O}(n \\log n)`
    """

    @staticmethod
    def transform_n(n: np.ndarray) -> np.ndarray:
        return np.vstack((np.ones(len(n)), n * np.log2(n))).T


linearithmic = _linearithmic()


class _quadratic(complexity):
    """
    Complexity class for quadratic time: :math:`\\mathcal{O}(n^2)`
    """

    @staticmethod
    def transform_n(n: np.ndarray) -> np.ndarray:
        return np.vstack((np.ones(len(n)), n * n)).T


quadratic = _quadratic()


class _cubic(complexity):
    """
    Complexity class for cubic time: :math:`\\mathcal{O}(n^3)`
    """

    @staticmethod
    def transform_n(n: np.ndarray) -> np.ndarray:
        return np.vstack((np.ones(len(n)), n * n * n)).T


cubic = _cubic()


class _exponential(complexity):
    """
    Complexity class for exponential time: :math:`\\mathcal{O}(2^n)`
    """

    @staticmethod
    def transform_n(n: np.ndarray) -> np.ndarray:
        return np.vstack((np.ones(len(n)), n)).T

    @staticmethod
    def transform_t(t: np.ndarray) -> np.ndarray:
        return np.log2(t)


exponential = _exponential()


complexity_classes = [constant, logarithmic, linear, linearithmic, quadratic, cubic]#, exponential]


class ComplexityUnion:
    """
    A complexity class that represents a union of multiple complexity classes.

    Complexity unions can be used to write annotations but cannot be used as classes to determine
    the complexity of a block of code; they are simply collections of acceptable complexity classes.

    This class does not ensure the uniqueness of the complexity classes it tracks; that is, if the
    same complexity class is added twice, it will be tracked twice.

    Args:
        *complexity_classes (:py:class:`complexity`): the complexity classes in the union
    """

    _complexity_classes = List[complexity]
    """the complexity classes contained in the union"""

    def __init__(self, *complexity_classes: complexity):
        self._complexity_classes = list(complexity_classes)

    def __or__(self, other: Any) -> "ComplexityUnion":
        """
        Create a complexity union through use of the ``|`` operator.

        Args:
            other (any): the other object in the union

        Returns:
            :py:class:`ComplexityUnion`: the union
        """
        return self.from_or(self, other)

    @classmethod
    def from_or(
        cls,
        left: Union[complexity, "ComplexityUnion"],
        right: Union[complexity, "ComplexityUnion"],
    ) -> "ComplexityUnion":
        """
        Create a complexity union from two operands.

        Args:
            left (:py:class:`complexity` or :py:class:`ComplexityUnion`): the left operand
            right (:py:class:`complexity` or :py:class:`ComplexityUnion`): the right operand

        Returns:
            :py:class:`ComplexityUnion`: the union containing both operands

        Raises:
            ``TypeError``: if either operand is not a complexity class or complexity union
        """
        if not isinstance(left, (complexity, cls)):
            raise TypeError(f"Attempted to combine a complexity class with an object of type {type(left)}")

        if not isinstance(right, (complexity, cls)):
            raise TypeError(f"Attempted to combine a complexity class with an object of type {type(right)}")

        if isinstance(left, cls):
            right_complexities = [right]
            if isinstance(right, cls):
                right_complexities = right.get_complexities()
            return cls(*left.get_complexities(), *right_complexities)

        elif isinstance(right, cls):
            return cls(left, *right.get_complexities())

        return cls(left, right)

    def add_complexity(self, complexity_class: complexity) -> None:
        """
        Add another complexity class to this union.

        Args:
            complexity_class (:py:class:`complexity`): the complexity class to add
        """
        self._complexity_classes.append(complexity_class)

    def get_complexities(self) -> List[complexity]:
        """
        Return a list of the complexity classes in this union.

        Returns:
            ``list[complexity]``: the complexity classes
        """
        return [*self._complexity_classes]

    def __eq__(self, other: Any) -> bool:
        """
        Determine whether another object is equal to this complexity union.

        An object is equal to a complexity union if it is also a complexity union and contains the
        same complexity classes.

        Returns:
            ``bool``: whether the other object is equal to this one
        """
        return isinstance(other, type(self)) and \
            set(self.get_complexities()) == set(other.get_complexities())



""""""

import random
import nbformat

from textwrap import dedent

from pybryt.preprocessors import IntermediateVariablePreprocessor


def test_preprocessor():
    """
    """
    nb = nbformat.v4.new_notebook()
    nb.cells.append(nbformat.v4.new_code_cell(dedent("""\
        a = True
        b = False
        f = lambda x: not x

        g = f(a) + f(b)

        if f(a) and f(b):
            print("hi")

        if f(a) or f(b):
            print("hi")
        
        if a or b:
            print("bye")

        l = [f(i) for i in [a, b]]

        f = lambda x: [not i for i in l]
        l = [a, b]
        if all(f(l)):
            print("ok")
        else:
            l = any(f(l))
    """)))

    ivp = IntermediateVariablePreprocessor()

    random.seed(42)
    nb = ivp.preprocess(nb)
    print(nb.cells[0].source)
    assert len(nb.cells) == 1
    assert nb.cells[0].source.strip() == dedent("""\
        a = True
        b = False
        f = (lambda x: (not x))
        var_HBRPOI = f(a)
        var_G8F1CB = f(b)
        g = (var_HBRPOI + var_G8F1CB)
        var_FNO6B9 = f(a)
        if (var_FNO6B9):
            var_M80O2R = f(b)
        if (var_FNO6B9 and var_M80O2R):
            var_AK1VRJ = print('hi')
            var_AK1VRJ
        var_NVGFYG = f(a)
        if (not (var_NVGFYG)):
            var_WWQC38 = f(b)
        if (var_NVGFYG or var_WWQC38):
            var_HYF9SX = print('hi')
            var_HYF9SX
        if (a or b):
            var_MECOSF = print('bye')
            var_MECOSF
        l = [f(i) for i in [a, b]]
        f = (lambda x: [(not i) for i in l])
        l = [a, b]
        var_KXWNRE = f(l)
        var_K8PK3Y = all(var_KXWNRE)
        if var_K8PK3Y:
            var_R9OUDO = print('ok')
            var_R9OUDO
        else:
            var_CUZREN = f(l)
            l = any(var_CUZREN)
    """).strip()



"""Tests for type annotations"""

import numpy as np
import pytest

from unittest import mock

import pybryt

from .utils import assert_object_attrs, generate_memory_footprint


def test_forbid_type():
    """
    """
    footprint = generate_memory_footprint()
    pybryt.Annotation.reset_tracked_annotations()

    a = pybryt.ForbidType(bool)
    res = a.check(footprint)

    assert_object_attrs(a, {"children__len": 0})
    assert_object_attrs(res, {
        "children": [],
        "satisfied": True,
        "_satisfied": True,
        "annotation": a,
        "timestamp": -1,
        "value": None,
    })

    assert a.to_dict() == {
        "name": "Annotation 1",
        "children": [],
        "success_message": None,
        "failure_message": None,
        "limit": None,
        "group": None,
        "type": "forbidtype",
        "type_": "<class 'bool'>",
    }

    a = pybryt.ForbidType(np.ndarray)
    res = a.check(footprint)
    assert_object_attrs(res, {
        "children": [],
        "satisfied": False,
        "_satisfied": False,
        "annotation": a,
        "timestamp": -1,
        "value": None,
    })

    # check constructor errors
    with pytest.raises(TypeError, match=f"1 is not a type"):
        pybryt.ForbidType(1)

    with mock.patch("pybryt.annotations.type_.dill") as mocked_dill:
        mocked_dill.dumps.side_effect = Exception()

        with pytest.raises(ValueError, match="Types must be serializable but the following error was thrown during serialization:\n"):
            pybryt.ForbidType(np.ndarray)



""""""

import os
import nbformat
import pkg_resources
import pytest
import tempfile

from copy import deepcopy
from functools import lru_cache
from textwrap import dedent
from unittest import mock

from pybryt import (
    check, generate_student_impls, ReferenceImplementation, ReferenceResult, StudentImplementation)
from pybryt.execution.memory_footprint import MemoryFootprint

from .test_reference import generate_reference_notebook


__PYBRYT_TRACING__ = False


def generate_student_notebook():
    """
    """
    nb = nbformat.v4.new_notebook()
    nb.cells.append(nbformat.v4.new_code_cell(dedent("""\
        import pybryt
    """)))
    nb.cells.append(nbformat.v4.new_code_cell(dedent("""\
        def median(S):
            sorted_S = sorted(S)
            size_of_set = len(S)
            middle = size_of_set // 2
            is_set_size_even = (size_of_set % 2) == 0
            if is_set_size_even:
                return (sorted_S[middle-1] + sorted_S[middle]) / 2
            else:
                return sorted_S[middle]
    """)))
    nb.cells.append(nbformat.v4.new_code_cell(dedent("""\
        import numpy as np
        np.random.seed(42)
        for _ in range(10):
            vals = [np.random.randint(-1000, 1000) for _ in range(np.random.randint(1, 1000))]
            val = median(vals)
    """)))
    return nb


@lru_cache(1)
def _generate_impl_cached():
    nb = generate_student_notebook()
    return nb, StudentImplementation(nb)

def generate_impl():
    return deepcopy(_generate_impl_cached())


def test_constructor():
    """
    """
    nb, stu = generate_impl()
    assert stu.nb is nb
    assert isinstance(stu.footprint, MemoryFootprint)
    assert len(stu.footprint) == 993

    with mock.patch("pybryt.student.execute_notebook") as mocked_exec:
        mocked_exec.return_value = MemoryFootprint()
        mocked_exec.return_value.set_executed_notebook(nb)

        with tempfile.NamedTemporaryFile(mode="w+", suffix=".ipynb") as ntf:
            nbformat.write(nb, ntf.name)

            stu = StudentImplementation(ntf.name)
            assert stu.footprint.num_steps == -1
            assert list(stu.footprint) == []
            assert stu.footprint.calls == []
            assert stu.nb == nb

            with tempfile.NamedTemporaryFile(mode="w+", suffix=".ipynb") as output_ntf:
                stu = StudentImplementation(ntf.name, output=output_ntf.name)
                assert nbformat.read(output_ntf.name, as_version=nbformat.NO_CONVERT) == nb

    with pytest.raises(TypeError, match="path_or_nb is of unsupported type <class 'int'>"):
        StudentImplementation(1)


def test_load_and_dump():
    """
    """
    _, stu = generate_impl()
    with tempfile.NamedTemporaryFile() as ntf:
        stu.dump(ntf.name)
        stu2 = StudentImplementation.load(ntf.name)

        assert len(stu.footprint) == len(stu2.footprint)
        assert stu.footprint.num_steps == stu2.footprint.num_steps

    enc_stu = stu.dumps()
    stu2 = StudentImplementation.loads(enc_stu)
    assert len(stu.footprint) == len(stu2.footprint)
    assert stu.footprint.num_steps == stu2.footprint.num_steps


def test_check():
    """
    """
    ref = ReferenceImplementation.compile(generate_reference_notebook(), name="foo")
    nb, stu = generate_impl()

    res = stu.check(ref)
    assert isinstance(res, ReferenceResult)

    res = stu.check([ref])
    assert isinstance(res, list) and len(res) == 1 and isinstance(res[0], ReferenceResult)

    with pytest.raises(TypeError, match="check cannot take values of type <class 'int'>"):
        stu.check(1)


def test_errors():
    nb = nbformat.v4.new_notebook()
    nb.cells.append(nbformat.v4.new_code_cell("raise Exception()"))

    with pytest.warns(UserWarning, match="Executing student notebook produced errors in the notebook"):
        stu = StudentImplementation(nb)

    assert len(stu.errors) == 1
    assert stu.errors[0]["ename"] == "Exception"
    assert stu.errors[0]["evalue"] == ""
    assert stu.errors[0]["output_type"] == "error"
    assert isinstance(stu.errors[0]["traceback"], list)
    assert len(stu.errors[0]) == 4
    

def test_check_cm(capsys):
    """
    """
    ref = ReferenceImplementation.compile(generate_reference_notebook(), name="foo")
    _, stu = generate_impl()

    with mock.patch.object(check, "_cache_check") as mocked_cache:
        with mock.patch("pybryt.student.FrameTracer") as mocked_frame_tracer:
            mocked_frame_tracer.return_value.get_footprint.return_value = stu.footprint
            check_cm = check(ref, cache=False)
            with check_cm:
                pass

            mocked_cache.assert_not_called()
            mocked_frame_tracer.return_value.start_trace.assert_called()
            mocked_frame_tracer.return_value.end_trace.assert_called()

        captured = capsys.readouterr()
        expected = dedent("""\
            REFERENCE: foo
            SATISFIED: True
            MESSAGES:
              - SUCCESS: Sorted the sample correctly
              - SUCCESS: Computed the size of the sample
              - SUCCESS: computed the correct median
        """)
        assert captured.out == expected

        with mock.patch("pybryt.student.FrameTracer") as mocked_frame_tracer:
            mocked_frame_tracer.return_value.get_footprint.return_value = stu.footprint
            ref_filename = pkg_resources.resource_filename(__name__, os.path.join("files", "expected_ref.pkl"))
            check_cm = check(ref_filename)
            with check_cm:
                pass

            mocked_cache.assert_called()

            check_cm2 = check([ref_filename])
            assert check_cm._ref == check_cm2._ref

        captured = capsys.readouterr()
        expected = dedent("""\
            REFERENCE: foo
            SATISFIED: True
            MESSAGES:
              - SUCCESS: Sorted the sample correctly
              - SUCCESS: Computed the size of the sample
              - SUCCESS: computed the correct median
        """)
        assert captured.out == expected

        # test errors
        with pytest.raises(ValueError, match="Cannot check against an empty list of references"):
            check([])

        with pytest.raises(TypeError, match="Invalid values in the reference list"):
            check([ref, "path", 1])

        # check by annotation group
        with mock.patch.object(StudentImplementation, "from_footprint") as mocked_ff, \
                mock.patch("pybryt.student.FrameTracer"), \
                mock.patch("pybryt.student.generate_report"):
            ref = ReferenceImplementation("groups", [])
            for run_group in ["1", "2", None]:
                with check(ref, group=run_group):
                    pass

                mocked_ff.return_value.check.assert_called_with([ref], group=run_group)

    # check caching
    with mock.patch("pybryt.student.FrameTracer") as mocked_frame_tracer:
        with mock.patch("pybryt.student.StudentImplementation") as mocked_stu, \
                mock.patch("pybryt.student.generate_report") as mocked_generate, \
                mock.patch("pybryt.student.os.makedirs") as mocked_makedirs:
            mocked_stu.from_footprint.return_value.check.return_value = [mock.MagicMock()]
            mocked_stu.from_footprint.return_value.check.return_value[0].name = "foo"

            check_cm = check(ref)
            with check_cm:
                check_cm._footprint = stu.footprint
            
            mocked_makedirs.assert_called_with(".pybryt_cache", exist_ok=True)
            mocked_stu.from_footprint.return_value.dump.assert_called()
            mocked_stu.from_footprint.return_value.check.return_value[0].dump.assert_called_with(".pybryt_cache/foo_results.pkl")


def test_from_cache():
    """
    """
    with mock.patch("pybryt.student.glob") as mocked_glob, \
            mock.patch.object(StudentImplementation, "load") as mocked_load, \
            mock.patch.object(StudentImplementation, "combine") as mocked_combine:
        mocked_glob.return_value = [".pybryt_cache/student_impl_foo.pkl", ".pybryt_cache/student_impl_bar.pkl"]
        StudentImplementation.from_cache(combine=False)

        mocked_load.assert_has_calls([mock.call(fp) for fp in mocked_glob.return_value])
        mocked_combine.assert_not_called()

        StudentImplementation.from_cache()

        mocked_combine.assert_called()


def test_combine():
    """
    """
    _, stu = generate_impl()
    stu2 = deepcopy(stu)
    stu2.footprint.add_value([1, 2, 3, 4], stu2.footprint.num_steps + 1)

    comb = StudentImplementation.combine([stu, stu2])
    assert len(comb.footprint) == len(stu.footprint) + 1
    assert comb.footprint.num_steps == stu.footprint.num_steps + stu2.footprint.num_steps
    assert comb.footprint.get_value(-1).timestamp == stu.footprint.num_steps + stu2.footprint.num_steps


def test_generate_student_impls():
    """
    """
    num_notebooks = 6
    nb, stu = generate_impl()
    nbs = [nb] * num_notebooks

    with mock.patch("pybryt.student.execute_notebook") as mocked_execute:
        mocked_execute.return_value = deepcopy(stu.footprint)
        stus = generate_student_impls(nbs)
    
    assert all(s == stu for s in stus)

    with mock.patch("pybryt.student.Process") as mocked_process:

        class MockedQueue:
            def __init__(self, *args, **kwargs):
                self.calls = 0

            def empty(self, *args, **kwargs):
                if self.calls >= num_notebooks:
                    return True
                self.calls += 1
                return False

            def get(self, *args, **kwargs):
                return (nb, stu)

        with mock.patch("pybryt.student.Queue") as mocked_queue:
            mocked_queue.return_value = mock.MagicMock(wraps=MockedQueue())
            stus = generate_student_impls(nbs, parallel=True)

    assert all(s == stu for s in stus)



"""Preprocessor for collecting the set of imported modules"""

import ast
import nbformat

from typing import Set

from .abstract_preprocessor import AbstractPreprocessor


class ImportFinder(ast.NodeVisitor):
    """
    An AST walker for collecting imported modules.
    """

    def __init__(self) -> None:
        self.imports = set()

    def visit_Import(self, node: ast.AST) -> None:
        """
        Visit an import statement and collect the names of the imported modules
        """
        for alias in node.names:
            self.imports.add(alias.name.split(".")[0])

    def visit_ImportFrom(self, node: ast.AST) -> None:
        """
        Visit an import from statement and collect the names of the imported modules
        """
        self.imports.add(node.module)


class ImportFindingPreprocessor(AbstractPreprocessor):
    """
    A notebook preprocessor that collects the set of modules import in the notebook.
    """

    imports: Set[str]

    def __init__(self) -> None:
        self.imports = set()
        super().__init__()

    def preprocess(self, nb: nbformat.NotebookNode) -> nbformat.NotebookNode:
        """
        Populate ``self.imports`` by walking the AST of each code cell.

        Args:
            nb (``nbformat.NotebookNode``): the notebook to be preprocessed

        Returns:
            ``nbformat.NotebookNode``: the notebook, unchanged
        """
        for cell in nb['cells']:
            if cell['cell_type'] == 'code':
                code = cell['source']
                code = self.transformer_manager.transform_cell(code)
                tree = ast.parse(code)
                import_finder = ImportFinder()
                import_finder.visit(tree)
                self.imports.update(import_finder.imports)

        return nb



"""Debug mode"""

__all__ = ["debug_mode", "disable_debug_mode", "enable_debug_mode"]

from contextlib import contextmanager
from typing import NoReturn


_DEBUG_MODE_ENABLED = False


def _debug_mode_enabled() -> bool:
    """
    Return whether debug mode is currently enabled.

    Returns:
        ``bool``: whether debug mode is enabled
    """
    return _DEBUG_MODE_ENABLED


def enable_debug_mode() -> None:
    """
    Enable PyBryt's debug mode.
    """
    global _DEBUG_MODE_ENABLED
    _DEBUG_MODE_ENABLED = True


def disable_debug_mode() -> None:
    """
    Disable PyBryt's debug mode.
    """
    global _DEBUG_MODE_ENABLED
    _DEBUG_MODE_ENABLED = False


@contextmanager
def debug_mode() -> None:
    """
    A context in which debug mode is enabled.

    When the context exits, debug mode is disabled globally. This means debug mode will be disabled
    even if it was enabled before the context was entered.
    """
    enable_debug_mode()
    yield
    disable_debug_mode()



"""Complexity analysis internals"""

from collections.abc import Sized
from typing import Optional, Union

from .memory_footprint import MemoryFootprint


COMPLEXITY_TRACING_ENABLED = False


class TimeComplexityResult:
    """
    A class for collecting the results of time complexity check blocks.

    Args:
        name (``str``): the name of the block
        n (``int``): the input length
        start (``int``): the step counter value at the start of the block
        stop (``int``): the step counter value at the end of the block
    """

    name: str
    """the name of the block"""

    n: int
    """the input length or the input itself"""

    start: int
    """the step counter value at the start of the block"""

    stop: int
    """the step counter value at the end of the block"""

    def __init__(self, name, n, start, stop):
        self.name = name
        self.n = n
        self.start = start
        self.stop = stop


class check_time_complexity:
    """
    A context manager for checking the time complexity of student code.

    Halts tracking of values in memory and sets the trace function to only increment the step
    counter. When the block exits, the step counter is checked an a ``TimeComplexityResult`` object
    is appended to the student's memory footprint.

    If the current call stack is not being traced by PyBryt, no action is taken.

    Args:
        name (``str``): the name of the check; should match with the name of an annotation in the
            reference implementation
        n (``Union[int, float, Sized]``): the input length or the input itself if it supports ``len``
    """

    name: str

    n: int

    start_steps: Optional[int]

    footprint: Optional[MemoryFootprint]

    def __init__(self, name: str, n: Union[int, float, Sized]):
        if isinstance(n, float):
            n = int(n)
        if isinstance(n, Sized):
            n = len(n)
        if not isinstance(n, int):
            try:
                n = int(n)
            except:
                raise TypeError(f"n has invalid type {type(n)}")
  
        self.name = name
        self.n = n
        self.start_steps, self.footprint = None, None

    def __enter__(self):
        global COMPLEXITY_TRACING_ENABLED

        if get_active_footprint() is not None:
            self.footprint = get_active_footprint()
            self.start_steps = self.footprint.counter.get_value()

        COMPLEXITY_TRACING_ENABLED = True

    def __exit__(self, exc_type, exc_value, traceback):
        global COMPLEXITY_TRACING_ENABLED

        COMPLEXITY_TRACING_ENABLED = False

        if self.start_steps is not None:
            end_steps = self.footprint.counter.get_value()
            self.footprint.add_value(TimeComplexityResult(
                self.name, self.n, self.start_steps, end_steps), allow_duplicates=True)

        return False


def is_complexity_tracing_enabled() -> bool:
    """
    Return whether complexity tracing is currently enabled.

    Returns:
        ``bool``: whether complexity tracing is currently enabled
    """
    return COMPLEXITY_TRACING_ENABLED


from .tracing import get_active_footprint



"""Tests for import annotations"""

import pytest

from unittest import mock

import pybryt
from pybryt.execution.memory_footprint import MemoryFootprint

from .utils import assert_object_attrs


IMPORTED_MODULES = ["numpy", "pandas", "matplotlib"]


def generate_import_mfp():
    footprint = MemoryFootprint()
    footprint.add_imports(*IMPORTED_MODULES)
    return footprint


@mock.patch.multiple(pybryt.ImportAnnotation, __abstractmethods__=frozenset())
def test_import_init_behaviors():
    """
    """
    with mock.patch("pybryt.annotations.import_.importlib") as mocked_importlib, \
            mock.patch("pybryt.annotations.import_.sys") as mocked_sys:
        pybryt.ImportAnnotation("foo")
        mocked_importlib.import_module.assert_called_with("foo")
        mocked_sys.modules.pop.assert_called_with("foo")

        mocked_sys.modules = {"foo": None}
        pybryt.ImportAnnotation("foo")
        assert mocked_sys.modules == {"foo": None}

        with pytest.raises(TypeError):
            pybryt.ImportAnnotation(1)

        mocked_importlib.import_module.side_effect = ImportError()
        with pytest.raises(ValueError):
            pybryt.ImportAnnotation("foo")


def test_require_import():
    """
    """
    footprint = generate_import_mfp()
    pybryt.Annotation.reset_tracked_annotations()

    with mock.patch("pybryt.annotations.import_.importlib") as mocked_importlib, \
            mock.patch("pybryt.annotations.import_.sys") as mocked_sys:
        a = pybryt.RequireImport(IMPORTED_MODULES[0])
        b = pybryt.RequireImport("foo")

    res1 = a.check(footprint)

    assert_object_attrs(a, {"children__len": 0})
    assert_object_attrs(res1, {
        "children": [],
        "satisfied": True,
        "_satisfied": True,
        "annotation": a,
        "timestamp": -1,
        "value": None,
    })

    assert a.to_dict() == {
        "name": "Annotation 1",
        "children": [],
        "success_message": None,
        "failure_message": None,
        "limit": None,
        "group": None,
        "type": "requireimport",
        "module": IMPORTED_MODULES[0],
    }

    res2 = b.check(footprint)
    assert not res2.satisfied



def test_forbid_import():
    """
    """
    footprint = generate_import_mfp()
    pybryt.Annotation.reset_tracked_annotations()

    with mock.patch("pybryt.annotations.import_.importlib") as mocked_importlib, \
            mock.patch("pybryt.annotations.import_.sys") as mocked_sys:
        a = pybryt.ForbidImport("foo")
        b = pybryt.ForbidImport(IMPORTED_MODULES[0])

    res1 = a.check(footprint)

    assert_object_attrs(a, {"children__len": 0})
    assert_object_attrs(res1, {
        "children": [],
        "satisfied": True,
        "_satisfied": True,
        "annotation": a,
        "timestamp": -1,
        "value": None,
    })

    assert a.to_dict() == {
        "name": "Annotation 1",
        "children": [],
        "success_message": None,
        "failure_message": None,
        "limit": None,
        "group": None,
        "type": "forbidimport",
        "module": "foo",
    }

    res2 = b.check(footprint)
    assert not res2.satisfied



"""Tests for relational annotations"""

import pytest

import pybryt

from .utils import assert_object_attrs, generate_memory_footprint


def test_before_annotation():
    """
    """
    footprint = generate_memory_footprint()
    pybryt.Annotation.reset_tracked_annotations()

    val1 = footprint.get_value(0).value
    val2, ts2 = footprint.get_value(1).value, footprint.get_value(1).timestamp

    v1 = pybryt.Value(val1)
    v2 = pybryt.Value(val2)
    
    v = v1.before(v2)
    res = v.check(footprint)

    # check attributes of BeforeAnnotation and AnnotationResult
    assert_object_attrs(v, {"children": (v1, v2)})
    assert_object_attrs(res, {
        "children__len": 2,
        "satisfied": True,
        "_satisfied": True,
        "annotation": v,
        "timestamp": -1,
        "satisfied_at": ts2,
    })

    assert v.to_dict() == {
        'name': 'Annotation 3', 
        'group': None, 
        'limit': None, 
        'success_message': None, 
        'failure_message': None, 
        'children': [
            {
                'name': 'Annotation 1', 
                'group': None, 
                'limit': None, 
                'success_message': None, 
                'failure_message': None, 
                'children': [], 
                'type': 'value', 
                'invariants': [], 
                'atol': None,
                'rtol': None,
            }, {
                'name': 
                'Annotation 2', 
                'group': None, 
                'limit': None, 
                'success_message': None, 
                'failure_message': None, 
                'children': [], 
                'type': 'value', 
                'invariants': [], 
                'atol': None,
                'rtol': None,
            }
        ], 
        'type': 'before',
    }

    v = v2.after(v1)
    res = v.check(footprint)

    # check attributes of BeforeAnnotation and AnnotationResult
    assert_object_attrs(v, {"children": (v1, v2)})
    assert_object_attrs(res, {
        "children__len": 2,
        "satisfied": True,
        "_satisfied": True,
        "annotation": v,
        "timestamp": -1,
        "satisfied_at": ts2,
    })

    v = v1.after(v2)
    res = v.check(footprint)

    # check attributes of BeforeAnnotation and AnnotationResult
    assert_object_attrs(v, {"children": (v2, v1)})
    assert_object_attrs(res, {
        "children__len": 2,
        "satisfied": False,
        "_satisfied": False,
        "annotation": v,
        "timestamp": -1,
        "satisfied_at": -1,
    })


def test_and_annotation():
    """
    """
    footprint = generate_memory_footprint()

    val1 = footprint.get_value(0).value
    val2, ts2 = footprint.get_value(1).value, footprint.get_value(1).timestamp

    v1 = pybryt.Value(val1)
    v2 = pybryt.Value(val2)
    
    v = v1 & v2
    res = v.check(footprint)

    assert_object_attrs(v, {"children": (v1, v2)})
    assert_object_attrs(res, {
        "children__len": 2,
        "satisfied": True,
        "_satisfied": True,
        "annotation": v,
        "timestamp": -1,
        "satisfied_at": ts2,
    })

    v3 = pybryt.Value([])
    v = v1 & v3
    res = v.check(footprint)

    assert_object_attrs(v, {"children": (v1, v3)})
    assert_object_attrs(res, {
        "children__len": 2,
        "satisfied": False,
        "_satisfied": False,
        "annotation": v,
        "timestamp": -1,
        "satisfied_at": -1,
    })

    v4 = pybryt.Value(6.02e+23)
    v = v4 & v2
    res = v.check(footprint)

    assert_object_attrs(v, {"children": (v4, v2)})
    assert_object_attrs(res, {
        "children__len": 2,
        "satisfied": False,
        "_satisfied": False,
        "annotation": v,
        "timestamp": -1,
        "satisfied_at": -1,
    })

    v = v4 & v3
    res = v.check(footprint)

    assert_object_attrs(v, {"children": (v4, v3)})
    assert_object_attrs(res, {
        "children__len": 2,
        "satisfied": False,
        "_satisfied": False,
        "annotation": v,
        "timestamp": -1,
        "satisfied_at": -1,
    })


def test_or_annotation():
    """
    """
    footprint = generate_memory_footprint()

    val1, ts1 = footprint.get_value(0).value, footprint.get_value(0).timestamp
    val2, ts2 = footprint.get_value(1).value, footprint.get_value(1).timestamp

    v1 = pybryt.Value(val1)
    v2 = pybryt.Value(val2)
    
    v = v1 | v2
    res = v.check(footprint)

    assert_object_attrs(v, {"children": (v1, v2)})
    assert_object_attrs(res, {
        "children__len": 2,
        "satisfied": True,
        "_satisfied": True,
        "annotation": v,
        "timestamp": -1,
        "satisfied_at": ts2,
    })

    v3 = pybryt.Value([])
    v = v1 | v3
    res = v.check(footprint)

    assert_object_attrs(v, {"children": (v1, v3)})
    assert_object_attrs(res, {
        "children__len": 2,
        "satisfied": True,
        "_satisfied": True,
        "annotation": v,
        "timestamp": -1,
        "satisfied_at": ts1,
    })

    v4 = pybryt.Value(6.02e+23)
    v = v4 | v2
    res = v.check(footprint)

    assert_object_attrs(v, {"children": (v4, v2)})
    assert_object_attrs(res, {
        "children__len": 2,
        "satisfied": True,
        "_satisfied": True,
        "annotation": v,
        "timestamp": -1,
        "satisfied_at": ts2,
    })

    v = v4 | v3
    res = v.check(footprint)

    assert_object_attrs(v, {"children": (v4, v3)})
    assert_object_attrs(res, {
        "children__len": 2,
        "satisfied": False,
        "_satisfied": False,
        "annotation": v,
        "timestamp": -1,
        "satisfied_at": -1,
    })


def test_xor_annotation():
    """
    """
    footprint = generate_memory_footprint()

    val1, ts1 = footprint.get_value(0).value, footprint.get_value(0).timestamp
    val2, ts2 = footprint.get_value(1).value, footprint.get_value(1).timestamp

    v1 = pybryt.Value(val1)
    v2 = pybryt.Value(val2)
    
    v = v1 ^ v2
    res = v.check(footprint)

    assert_object_attrs(v, {"children": (v1, v2)})
    assert_object_attrs(res, {
        "children__len": 2,
        "satisfied": False,
        "_satisfied": False,
        "annotation": v,
        "timestamp": -1,
        "satisfied_at": -1,
    })

    v3 = pybryt.Value([])
    v = v1 ^ v3
    res = v.check(footprint)

    assert_object_attrs(v, {"children": (v1, v3)})
    assert_object_attrs(res, {
        "children__len": 2,
        "satisfied": True,
        "_satisfied": True,
        "annotation": v,
        "timestamp": -1,
        "satisfied_at": ts1,
    })

    v4 = pybryt.Value(6.02e+23)
    v = v4 ^ v2
    res = v.check(footprint)

    assert_object_attrs(v, {"children": (v4, v2)})
    assert_object_attrs(res, {
        "children__len": 2,
        "satisfied": True,
        "_satisfied": True,
        "annotation": v,
        "timestamp": -1,
        "satisfied_at": ts2,
    })

    v = v4 ^ v3
    res = v.check(footprint)

    assert_object_attrs(v, {"children": (v4, v3)})
    assert_object_attrs(res, {
        "children__len": 2,
        "satisfied": False,
        "_satisfied": False,
        "annotation": v,
        "timestamp": -1,
        "satisfied_at": -1,
    })


def test_not_annotation():
    """
    """
    footprint = generate_memory_footprint()

    val1 = footprint.get_value(0).value

    v1 = pybryt.Value(val1)
    v = ~v1
    res = v.check(footprint)

    assert_object_attrs(v, {"children": (v1, )})
    assert_object_attrs(res, {
        "children__len": 1,
        "satisfied": False,
        "_satisfied": False,
        "annotation": v,
        "timestamp": -1,
        "satisfied_at": -1,
    })

    v3 = pybryt.Value([])
    v = ~v3
    res = v.check(footprint)

    assert_object_attrs(v, {"children": (v3, )})
    assert_object_attrs(res, {
        "children__len": 1,
        "satisfied": True,
        "_satisfied": True,
        "annotation": v,
        "timestamp": -1,
        "satisfied_at": -1,
    })


def test_constructor_errors():
    """
    """
    with pytest.raises(ValueError):
        pybryt.AndAnnotation([pybryt.Value(1), "not an annotation"])



"""Tests for initial conditions"""

import numpy as np
import operator

from unittest import mock

import pytest

from pybryt import InitialCondition
from pybryt.execution.memory_footprint import MemoryFootprint, MemoryFootprintValue

from .utils import assert_object_attrs


def test_constructor():
    """
    Tests for the ``InitialCondition`` constructor.
    """
    name = "foo"
    transforms = [lambda x: 2 * x, lambda x: np.cos(x)]
    ic = InitialCondition(name, transforms)
    assert_object_attrs(ic, {
        "name": name,
        "transforms": transforms,
    })

    # test errors
    with pytest.raises(TypeError):
        InitialCondition(2)


def test_apply():
    """
    Tests for ``InitialCondition.apply``.
    """
    name = "foo"
    transforms = [lambda x: 2 * x, lambda x: np.cos(x)]
    ic = InitialCondition(name, transforms)

    new_transform = lambda x: np.sin(x)
    ic2 = ic.apply(new_transform)

    assert isinstance(ic2, InitialCondition) and ic2 is not ic
    assert_object_attrs(ic2, {
        "name": name,
        "transforms": transforms + [new_transform],
    })


def test_supply_value():
    """
    Tests for ``InitialCondition.supply_value``.
    """
    name = "foo"
    transforms = [mock.MagicMock() for _ in range(10)]
    ic = InitialCondition(name, transforms)

    val = 2
    res = ic.supply_value(val)

    expected_arg = val
    for t in transforms:
        t.assert_called_once_with(expected_arg)
        expected_arg = t.return_value

    assert res == transforms[-1].return_value


def test_supply_footprint_and_supply_values():
    """
    Tests for ``InitialCondition.supply_footprint`` and ``InitialCondition.supply_values``.
    """
    name = "foo"
    transforms = [mock.MagicMock() for _ in range(10)]
    ic = InitialCondition(name, transforms)

    val = 2
    mocked_fp = mock.MagicMock()
    mocked_fp.get_initial_conditions.return_value = {name: val}

    res = ic.supply_footprint(mocked_fp)

    mocked_fp.get_initial_conditions.assert_called_once()

    expected_arg = val
    for t in transforms:
        t.assert_called_once_with(expected_arg)
        expected_arg = t.return_value

    assert res == transforms[-1].return_value

    # test nested initial conditions
    n1, n2, v1, v2 = "foo", "bar", 2, 3
    ic1, ic2 = InitialCondition(n1), InitialCondition(n2)
    vals = {n1: v1, n2: v2}
    ic3 = ic1 + ic2
    assert ic3.supply_values(vals) == v1 + v2

    # test errors
    with pytest.raises(ValueError, match=f"The provided values do not have key '{name}'"):
        ic.supply_values({"bar": 1})


def test_eq():
    """
    Tests for ``==`` comparisons on ``InitialCondition``s.
    """
    ic1, ic2, ic3 = InitialCondition("foo"), InitialCondition("foo"), InitialCondition("bar")

    assert ic1 == ic2
    assert ic1 != ic3

    t = lambda v: 2 * v
    ic2 = ic2.apply(t)
    assert ic1 != ic2

    ic1 = ic1.apply(t)
    assert ic1 == ic2


def test_operators():
    """
    Tests for the use of operators on ``InitialCondition``s.
    """
    ic = InitialCondition("foo")
    l, r, m = 2, 3, 4

    def test_op(op, right, use_m=False):
        if right:
            if use_m:
                op_ic = op(l, ic, m)
            else:
                op_ic = op(l, ic)

        else:
            if use_m:
                op_ic = op(ic, r, m)
            else:
                op_ic = op(ic, r)

        if use_m:
            exp = op(l, r, m)
        else:
            exp = op(l, r)

        assert isinstance(op_ic, InitialCondition)

        eq = op_ic.supply_value(r if right else l) == exp
        if isinstance(exp, np.ndarray):
            assert eq.all()
        else:
            assert eq

    # test binary ops
    bin_ops = [
        operator.add, operator.sub, operator.mul, operator.truediv, operator.floordiv, operator.mod,\
        operator.pow, operator.lshift, operator.rshift, operator.and_, operator.xor, operator.or_,
        divmod,
    ]
    for op in bin_ops:
        test_op(op, False)
        test_op(op, True)

    # test ternary pow
    test_op(pow, False, use_m=True)

    # test unary ops
    unary_ops = [operator.neg, operator.pos, operator.abs, operator.invert]
    for op in unary_ops:
        op_ic = op(ic)
        assert isinstance(op_ic, InitialCondition)
        assert op_ic.supply_value(l) == op(l)

    # test matmul
    l, r = mock.MagicMock(), mock.MagicMock()
    test_op(operator.matmul, False)
    l.__matmul__.assert_called_with(r)

    l = mock.Mock()
    test_op(operator.matmul, True)
    r.__rmatmul__.assert_called_with(l)



"""Various utilities for PyBryt"""

import os
import json
import random
import base64
import string
import dill
import hashlib
import time
import nbformat

from abc import ABC, abstractmethod
from typing import Any, List, Optional, Union
from IPython import get_ipython
from IPython.display import publish_display_data


class UnpickleableError(Exception):
    """
    An exception to raise when :py:func:`pybryt.utils.pickle_and_hash` fails.
    """


def pickle_and_hash(obj: Any) -> str:
    """
    Uses ``dill`` to pickle an object and returns the SHA-512 hash of the returned bytes.

    Args:
        obj (``object``): the object to pickle and hash
    
    Returns:
        ``str``: the hex digest of the SHA-512 hash of the pickled object

    Raises:
        :py:class:`pybryt.utils.UnpickleableError`: if the object cannot be pickled
    """
    try:
        s = dill.dumps(obj)
    except:
        raise UnpickleableError()

    return hashlib.sha512(s).hexdigest()


def filter_pickleable_list(lst: List[Any]) -> None:
    """
    Removes all elements from a list that cannot be pickled with ``dill``.

    Args:
        lst (``list[object]``): the list to filter
    """
    to_delete = []
    for i, v in enumerate(lst):
        try:
            dill.dumps(v)
        except:
            to_delete.append(i)
    
    to_delete.reverse()
    for i in to_delete:
        lst.pop(i)


def notebook_to_string(nb_path: Union[str, nbformat.NotebookNode]) -> str:
    """
    Converts a Jupyter Notebook to a string of executable Python code.

    Args:
        nb_path (``str`` or ``nbformat.NotebookNode``): the path to the notebook or the notebook
            in-memory
    
    Returns:
        ``str``: the string of all Python code from the notebook
    """   
    if isinstance(nb_path, str):
        with open(nb_path) as f:
            nb = json.load(f)
    elif isinstance(nb_path, nbformat.NotebookNode):
        nb = nb_path
    else:
        raise TypeError("invalid notebook type")
    
    source = ""
    for cell in nb['cells']:
        if cell['cell_type'] == 'code':
            if isinstance(cell['source'], list):
                source += "".join(cell['source']) + "\n"
            else:
                assert isinstance(cell['source'], str), f"could not parse notebook cell: {cell}"
                source += cell['source'] + "\n"

    source = "\n".join(l for l in source.split("\n") if not l.startswith("%") and not l.startswith("!"))
    return source


def make_secret(size=6, chars=string.ascii_uppercase + string.digits):
    """
    This function generates a random string using the given length and character set.
    
    Args:
        size (``int``, optional): the length of output name
        chars (``str``, optional): the set of characters used to create function name
    
    Returns:
        ``str``: randomly-generated string
    """
    return ''.join(random.choice(chars) for _ in range(size))


def save_notebook(filename, timeout=10):
    """
    Force-saves a Jupyter Notebook by displaying JavaScript.

    Args:
        filename (``str``): path to notebook file being saved
        timeout (``int`` or ``float``): number of seconds to wait for save before timing-out
    
    Returns
        ``bool``: whether the notebook was saved successfully
    """
    if get_ipython() is not None:
        f = open(filename, "rb")
        md5 = hashlib.md5(f.read()).hexdigest()
        start = time.time_ns()
        publish_display_data({"application/javascript": """
            if (typeof Jupyter !== "undefined") {
                Jupyter.notebook.save_notebook();
            }
        """})
        
        curr = md5
        while curr == md5 and time.time_ns() - start <= timeout * 10**9:
            time.sleep(1)
            f.seek(0)
            curr = hashlib.md5(f.read()).hexdigest()
        
        f.close()

        return curr != md5


def get_stem(fp) -> str:
    """
    Returns the stem of a filepath.

    Args:
        fp (``str``): the filepath

    Returns:
        ``str``: the stem of the filepath
    """
    return os.path.splitext(os.path.split(fp)[1])[0]


class Serializable(ABC):
    """
    A class that implements serialization using the ``dill`` library.
    """

    @property
    @abstractmethod
    def _default_dump_dest(self) -> str:
        """
        Default destination path for the ``dump`` method
        """
        ... # pragma: no cover

    def dump(self, dest: Optional[str] = None) -> None:
        """
        Pickles this object to a file.

        Args:
            dest (``str``, optional): the path to the file
        """
        if dest is None:
            dest = self._default_dump_dest
        with open(dest, "wb+") as f:
            dill.dump(self, f)

    def dumps(self) -> str:
        """
        Pickles this object to a base-64-encoded string.

        Returns:
           ``str``: the pickled and encoded object
        """
        bits = dill.dumps(self)
        return base64.b64encode(bits).decode("ascii")

    @classmethod
    def load(cls, file: str) -> 'Serializable':
        """
        Unpickles an object from a file.

        Args:
            file (``str``): the path to the file
        
        Returns:
            ``Serializable``: the unpickled object
        """
        with open(file, "rb") as f:
            instance = dill.load(f)
        if not isinstance(instance, cls):
            raise TypeError(f"Unpickled object is not of type {cls}")
        return instance

    @classmethod
    def loads(cls, data: str) -> "Serializable":
        """
        Unpickles an object from a base-64-encoded string.

        Args:
            data (``str``): the pickled and encoded object
        
        Returns:
            ``Serializable``: the unpickled object
        """
        instance = dill.loads(base64.b64decode(data.encode("ascii")))
        if not isinstance(instance, cls):
            raise TypeError(f"Unpickled object is not of type {cls}")
        return instance



"""Tests for value annotation invariants"""

import numpy as np

from unittest import mock

import pybryt.invariants as inv


def test_call_structure():
    with mock.patch.object(inv.invariant, "run") as mocked_run:
        inv.invariant()
        mocked_run.assert_called()

    assert inv.invariant([]) is None


def test_string_capitalization():
    np.random.seed(42)
    values = ["someString", "some_otherO!I(I*$NDdnnfkf", 120484, ["hi", "there"], np.random.uniform(size=100)]
    expected_values = [s.lower() if isinstance(s, str) else s for s in values]
    np.testing.assert_equal(inv.string_capitalization(values), expected_values)


def test_matrix_transpose():
    np.random.seed(42)
    arr = np.random.uniform(size=(10,10))
    values = ["someString", "some_otherO!I(I*$NDdnnfkf", 120484, ["hi", "there"], arr, [[1, 2, 3], [4, 5, 6]]]
    print(inv.matrix_transpose(values))
    expected_values = ["someString", "some_otherO!I(I*$NDdnnfkf", 120484, ["hi", "there"], ["hi", "there"], \
        arr, arr.T, [[1, 2, 3], [4, 5, 6]], [[1, 4], [2, 5], [3, 6]]]
    np.testing.assert_equal(inv.matrix_transpose(values), expected_values)


def test_list_permutation():
    np.random.seed(42)
    arr = np.random.uniform(size=(10,10))
    values = ["someString", "some_otherO!I(I*$NDdnnfkf", 120484, ["there", "hi", "z"], arr, [[1, 2, 3], [4, 5, 6]]]
    expected_values = ["someString", "some_otherO!I(I*$NDdnnfkf", 120484, ["hi", "there", "z"], np.sort(arr),  \
        [[1, 2, 3], [4, 5, 6]]]
    np.testing.assert_equal(inv.list_permutation(values), expected_values)


def test_alias():
    import pybryt.annotations.invariants as inv2
    assert inv2.string_capitalization is inv.string_capitalization



"""Tests for PyBryt's Otter plugin"""

import os
import base64
import shutil
import tempfile
import nbformat
import pathlib
import dill
import pkg_resources
import pytest

from otter.assign.assignment import Assignment
from otter.test_files import GradingResults 
from textwrap import dedent
from unittest import mock

from pybryt import ReferenceImplementation
from pybryt.integrations.otter import OtterPlugin

from ..test_reference import generate_reference_notebook
from ..test_student import generate_student_notebook


def generate_assignment_config(master, result, **kwargs):
    a = Assignment()
    a.master = master
    a.result = result
    a.update(kwargs)
    return a


def test_during_assign():
    nb = generate_reference_notebook()
    with tempfile.NamedTemporaryFile(mode="w+", suffix=".ipynb") as ntf:
        nbformat.write(nb, ntf)
        ntf.seek(0)
        plg = OtterPlugin("", {}, {"references": [ntf.name]})

        rdir = pathlib.Path(tempfile.mkdtemp())
        os.makedirs(rdir / 'autograder')
        os.makedirs(rdir / 'student')

        assignment = generate_assignment_config("", rdir)

        plg.during_assign(assignment)

        ntf_stem = pathlib.Path(ntf.name).stem

        try:
            agref, stref = str((rdir / 'autograder' / ntf_stem).with_suffix(".pkl")), \
                str((rdir / 'student' / ntf_stem).with_suffix(".pkl"))
            # breakpoint()
            assert os.path.isfile(agref)
            assert os.path.isfile(stref)

            # refs = ReferenceImplementation.load(agref)
            with open(agref, "rb") as f:
                refs = dill.load(f)
            assert isinstance(refs, list)
            assert len(refs) == 1
            assert isinstance(refs[0], ReferenceImplementation)

            # refs = ReferenceImplementation.load(stref)
            with open(stref, "rb") as f:
                refs = dill.load(f)
            assert isinstance(refs, list)
            assert len(refs) == 1
            assert isinstance(refs[0], ReferenceImplementation)

        except:
            shutil.rmtree(rdir)
            raise

        else:
            shutil.rmtree(rdir)


def test_during_generate():
    nb = generate_reference_notebook()
    with tempfile.NamedTemporaryFile(mode="w+", suffix=".ipynb") as ntf:
        nbformat.write(nb, ntf)
        ntf.seek(0)

        otter_config = {
            "plugins": [
                {
                    OtterPlugin.IMPORTABLE_NAME: {
                        "references": [ntf.name]
                    }
                }
            ]
        }

        plg = OtterPlugin("", {}, {"references": [ntf.name]})
        plg.during_generate(otter_config, None)

        cfg = otter_config["plugins"][0][OtterPlugin.IMPORTABLE_NAME]
        assert "reference_bytes" in cfg

        refs = dill.loads(base64.b64decode(cfg["reference_bytes"]))
        assert isinstance(refs, list)
        assert len(refs) == 1
        assert isinstance(refs[0], ReferenceImplementation)

    # test after running with otter assign
    plg._cached_refs = refs
    with mock.patch("os.chdir"):
        plg.during_generate(otter_config, generate_assignment_config(pathlib.Path(os.getcwd()), ""))

        cfg = otter_config["plugins"][0][OtterPlugin.IMPORTABLE_NAME]
        assert "reference_bytes" in cfg

        refs = dill.loads(base64.b64decode(cfg["reference_bytes"]))
        assert isinstance(refs, list)
        assert len(refs) == 1
        assert isinstance(refs[0], ReferenceImplementation)


def test_from_notebook(capsys):
    nb = generate_student_notebook()

    with tempfile.NamedTemporaryFile(mode="w+", suffix=".ipynb") as ntf:
        nb.cells.append(nbformat.v4.new_code_cell(dedent(f"""\
            import otter
            grader = otter.Notebook("{ntf.name}")
            grader.run_plugin("pybryt.integrations.otter.OtterPlugin", "ref.sey")
            grader.add_plugin_files("pybryt.integrations.otter.OtterPlugin")
        """)))

        nbformat.write(nb, ntf)
        ntf.seek(0)

        ref_path = str(pathlib.Path(__file__).parent.parent / 'files' / 'expected_ref.pkl')

        plg = OtterPlugin(ntf.name, {}, {})
        with pytest.warns(UserWarning, match="ould not force-save notebook; the results of this "
                                "call will be based on the last saved version of this notebook."):
            plg.from_notebook(ref_path)

        captured = capsys.readouterr()
        expected = dedent("""\
            PyBryt Reference Messages:
            - SUCCESS: Sorted the sample correctly
            - SUCCESS: Computed the size of the sample
            - SUCCESS: computed the correct median

            REFERENCE SATISFIED
        """)
        assert captured.out == expected

    with tempfile.NamedTemporaryFile(mode="w+", suffix=".ipynb") as ntf:
        nbformat.write(nbformat.v4.new_notebook(), ntf)
        ntf.seek(0)

        plg = OtterPlugin(ntf.name, {}, {})

        with pytest.warns(UserWarning, match="ould not force-save notebook; the results of this "
                                "call will be based on the last saved version of this notebook."):
            plg.from_notebook(ref_path)

        captured = capsys.readouterr()
        expected = dedent("""\
            PyBryt Reference Messages:
            - ERROR: The sample was not sorted
            - ERROR: Did not capture the size of the set to determine if it is odd or even
            - ERROR: failed to compute the median

            NO REFERENCE SATISFIED
        """)
        assert captured.out == expected



"""Abstract base class for notebook preprocessors"""

import nbformat

from abc import ABC, abstractmethod
from IPython.core.inputtransformer2 import TransformerManager


class AbstractPreprocessor(ABC):
    """
    An abstract base class for notebook execution preprocessors.
    """

    transformer_manager: TransformerManager

    def __init__(self) -> None:
        self.transformer_manager = TransformerManager()

    @abstractmethod
    def preprocess(self, nb: nbformat.NotebookNode) -> nbformat.NotebookNode:
        """
        Preprocesses a notebook for execution.

        Args:
            nb (``nbformat.NotebookNode``): the notebook to be preprocessed

        Returns:
            ``nbformat.NotebookNode``: the updated notebook
        """
        ...  # pragma: no cover
